open HogeAsm

let string_of_reg x = (* % *)
  if not (is_reg x) then (prerr_endline x; assert false)
  else String.sub x 1 (String.length x - 1)
let regs = Array.map string_of_reg regs
let fregs = regs
let allregs = Array.to_list regs
let allfregs = allregs
let reg_sw = string_of_reg reg_sw
let reg_fsw = reg_sw
let reg_cl = string_of_reg reg_cl
let reg_ra = string_of_reg reg_ra
let reg_sp = string_of_reg reg_sp
let reg_hp = string_of_reg reg_hp

let rec remove_percent = function
  | Ans e -> Ans(sub1 e)
  | Let((x,t),e,rest) -> Let((sub2 x,t), sub1 e, remove_percent rest)
  | Forget(x,rest) -> Forget(sub2 x, remove_percent rest)
and sub1 = function
  | Nop | SetL _ | Restore _ | Imm _ as e -> e
  | Add(a,b) -> Add(sub2 a, sub3 b)
  | Sub(a,b) -> Sub(sub2 a, sub3 b)
  | SLL(a,i) -> SLL(sub2 a, i)
  | Ld(a,i) -> Ld(sub2 a, i)
  | St(a,b,i) -> St(sub2 a, sub2 b, i)
  | FNeg(a) -> FNeg(sub2 a)
  | FAdd(a,b) -> FAdd(sub2 a, sub2 b)
  | FSub(a,b) -> FSub(sub2 a, sub2 b)
  | FMul(a,b) -> FMul(sub2 a, sub2 b)
  | FDiv(a,b) -> FDiv(sub2 a, sub2 b)
  | IfEq(a,b,th,el) -> IfEq(sub2 a, sub3 b, remove_percent th, remove_percent el)
  | IfLE(a,b,th,el) -> IfLE(sub2 a, sub3 b, remove_percent th, remove_percent el)
  | IfGE(a,b,th,el) -> IfGE(sub2 a, sub3 b, remove_percent th, remove_percent el)
  | IfFEq(a,b,th,el) -> IfFEq(sub2 a, sub2 b, remove_percent th, remove_percent el)
  | IfFLE(a,b,th,el) -> IfFLE(sub2 a, sub2 b, remove_percent th, remove_percent el)
  | CallCls(a,bs,cs) -> CallCls(sub2 a, List.map sub2 bs, List.map sub2 cs)
  | CallDir(l,bs,cs) -> CallDir(l,      List.map sub2 bs, List.map sub2 cs)
  | Save(a,b) -> Save(sub2 a, b)
and sub2 = string_of_reg
and sub3 = function
  | V(a) -> V(sub2 a)
  | x -> x

external gethi : float -> int32 = "gethi"
external getlo : float -> int32 = "getlo"

let stackset = ref S.empty (* ǤSave줿ѿν (caml2html: emit_stackset) *)
let stackmap = ref [] (* Save줿ѿΡåˤ (caml2html: emit_stackmap) *)
let save x =
  stackset := S.add x !stackset;
  if not (List.mem x !stackmap) then
    stackmap := !stackmap @ [x]
let savef = save
let locate x =
  let rec loc = function
    | [] -> []
    | y :: zs when x = y -> 0 :: List.map succ (loc zs)
    | y :: zs -> List.map succ (loc zs) in
  loc !stackmap
let offset x = List.hd (locate x)
let stacksize () = List.length !stackmap + 1

let pp_id_or_imm = function
  | V(x) -> x
  | C(i) -> string_of_int i

(* ؿƤӽФΤ˰¤ؤ(register shuffling) (caml2html: emit_shuffle) *)
let rec shuffle sw xys =
  (* remove identical moves *)
  let _, xys = List.partition (fun (x, y) -> x = y) xys in
  (* find acyclic moves *)
  match List.partition (fun (_, y) -> List.mem_assoc y xys) xys with
  | [], [] -> []
  | (x, y) :: xys, [] -> (* no acyclic moves; resolve a cyclic move *)
      (y, sw) :: (x, y) :: shuffle sw (List.map
					 (function
					   | (y', z) when y = y' -> (sw, z)
					   | yz -> yz)
					 xys)
  | xys, acyc -> acyc @ shuffle sw xys

type dest = Tail | NonTail of Id.t (* ɤɽǡ (caml2html: emit_dest) *)
let rec g oc = function (* ̿Υ֥ (caml2html: emit_g) *)
  | dest, Ans(exp) -> g' oc (dest, exp)
  | dest, Let((x, t), exp, e) ->
      g' oc (NonTail(x), exp);
      g oc (dest, e)
  | _, Forget _ -> assert false
and g' oc = function (* ̿Υ֥ (caml2html: emit_gprime) *)
  (* Ǥʤä׻̤dest˥å (caml2html: emit_nontail) *)
  | NonTail(_), Nop -> Printf.fprintf oc "# this is debug message 1\n"
  | NonTail(x), SetL(Id.L(y)) ->
      Printf.fprintf oc "\taddi %s r0 hi(%s)\n" x y;
      Printf.fprintf oc "\tsll %s %s 16\n" x x;
      Printf.fprintf oc "\taddi %s %s lo(%s)\n" x x y;
  | NonTail(x), Imm(i32) ->
      let hi = Int32.to_int (Int32.shift_right i32 16) in
      let i = Int32.to_int i32 in
      let lo = (i lsl 15) asr 15 in (* int31bitsʤΤ15bitsեȤɤФ *)
      let hi = hi + ((i lsr 15) land 1) in
      if hi != 0 then begin
        Printf.fprintf oc "\taddi %s r0 %d\n" x hi;
        Printf.fprintf oc "\tsll %s %s 16\n" x x;
        if lo != 0 then Printf.fprintf oc "\taddi %s %s %d\n" x x lo;
      end else
        Printf.fprintf oc "\taddi %s r0 %d\n" x lo
  | NonTail(x), Add(y, V(z)) -> Printf.fprintf oc "\tadd %s %s %s\n" x y z
  | NonTail(x), Add(y, C(z)) -> Printf.fprintf oc "\taddi %s %s %d\n" x y z
  | NonTail(x), Sub(y, V(z)) -> Printf.fprintf oc "\tsub %s %s %s\n" x y z
  | NonTail(x), Sub(y, C(z)) -> Printf.fprintf oc "\tsubi %s %s %d\n" x y z
  | NonTail(x), SLL(y, z) -> Printf.fprintf oc "\tsll %s %s %d\n" x y z
  | NonTail(x), Ld(y, z) -> Printf.fprintf oc "\tload %s %s %d\n" x y z
  | NonTail(_), St(x, y, z) -> Printf.fprintf oc "\tstore %s %s %d\n" x y z
  | NonTail(x), FNeg(y) -> Printf.fprintf oc "\tfneg %s %s\n" x y;
  | NonTail(x), FAdd(y, z) -> Printf.fprintf oc "\tfadd %s %s %s\n"  x y z
  | NonTail(x), FSub(y, z) -> Printf.fprintf oc "\tfsub %s %s %s\n"  x y z
  | NonTail(x), FMul(y, z) -> Printf.fprintf oc "\tfmul %s %s %s\n"  x y z
  | NonTail(x), FDiv(y, z) -> Printf.fprintf oc "\tfdiv %s %s %s\n"  x y z
  (* β̿μ (caml2html: emit_save) *)
  | NonTail(_), Save(x, y) when List.mem x allregs && not (S.mem y !stackset) ->
      save y;
      Printf.fprintf oc "\tstore %s %s %d\n" x reg_sp (offset y)
  | NonTail(_), Save(x, y) -> assert (S.mem y !stackset); ()
  (* β̿μ (caml2html: emit_restore) *)
  | NonTail(x), Restore(y) ->
      assert (List.mem x allregs);
      Printf.fprintf oc "\tload %s %s %d\n" x reg_sp (offset y)
  (* ä׻̤쥸˥åȤret (caml2html: emit_tailret) *)
  | Tail, (Nop | St _ | Save _ as exp) ->
      g' oc (NonTail(Id.gentmp Type.Unit), exp);
      Printf.fprintf oc "\tjr r31\n"
  | Tail, (SetL _ | Add _ | Sub _ | SLL _ | Ld _ | Imm _ as exp) ->
      g' oc (NonTail(regs.(0)), exp);
      Printf.fprintf oc "\tjr r31\n"
  | Tail, (FNeg _ | FAdd _ | FSub _ | FMul _ | FDiv _ as exp) ->
      g' oc (NonTail(fregs.(0)), exp);
      Printf.fprintf oc "\tjr r31\n"
  | Tail, (Restore(x) as exp) ->
      (match locate x with
      | [i] -> g' oc (NonTail(regs.(0)), exp)
      | _ -> assert false);
      Printf.fprintf oc "\tjr r31\n";
  (* ʬ *)
  | Tail, IfEq(x, V(y), e1, e2) ->
      let label = Id.genid ("be_else") in
      Printf.fprintf oc "\tbne %s %s %s\n" x y label;
      g'_tail_if oc e1 e2 label
  | Tail, IfLE(x, V(y), e1, e2) ->
      let label = Id.genid ("ble_else") in
      Printf.fprintf oc "\tbg %s %s %s\n" x y label;
      g'_tail_if oc e1 e2 label
  | Tail, IfGE(x, V(y), e1, e2) ->
      let label = Id.genid ("bge_else") in
      Printf.fprintf oc "\tbl %s %s %s\n" x y label;
      g'_tail_if oc e1 e2 label
  | Tail, IfFEq(x, y, e1, e2) ->
      let label = Id.genid ("fbe_else") in
      Printf.fprintf oc "\tfbne %s %s %s\n" x y label;
      g'_tail_if oc e1 e2 label
  | Tail, IfFLE(x, y, e1, e2) ->
      let label = Id.genid ("fble_else") in
      Printf.fprintf oc "\tfbg %s %s %s\n" x y label;
      g'_tail_if oc e1 e2 label
  | dest, IfEq(x, V(y), e1, e2) ->
      let label = Id.genid "be_else" in
      Printf.fprintf oc "\tbne %s %s %s\n" x y label;
      g'_non_tail_if oc dest e1 e2 label (Id.genid "be_cont")
  | dest, IfLE(x, V(y), e1, e2) ->
      let label = Id.genid "ble_else" in
      Printf.fprintf oc "\tbg %s %s %s\n" x y label;
      g'_non_tail_if oc dest e1 e2 label (Id.genid "ble_cont")
  | dest, IfGE(x, V(y), e1, e2) ->
      let label = Id.genid ("bge_else") in
      Printf.fprintf oc "\tbl %s %s %s\n" x y label;
      g'_non_tail_if oc dest e1 e2 label (Id.genid "bge_cont")
  | dest, IfFEq(x, y, e1, e2) ->
      let label = Id.genid ("fbe_else") in
      Printf.fprintf oc "\tfbne %s %s %s\n" x y label;
      g'_non_tail_if oc dest e1 e2 label (Id.genid "fbe_cont")
  | dest, IfFLE(x, y, e1, e2) ->
      let label = Id.genid ("fbg_else") in
      Printf.fprintf oc "\tfbg %s %s %s\n" x y label;
      g'_non_tail_if oc dest e1 e2 label (Id.genid "fble_cont")
  (* ؿƤӽФβ̿μ (caml2html: emit_call) *)
  | Tail, CallCls(x, ys, zs) -> (* ƤӽФ (caml2html: emit_tailcall) *)
      g'_args oc [(x, reg_cl)] ys zs;
      Printf.fprintf oc "\tload %s %s 0\n" reg_sw reg_cl;
      Printf.fprintf oc "\tjr %s\n" reg_sw;
  | Tail, CallDir(Id.L(x), ys, zs) -> (* ƤӽФ *)
      g'_args oc [] ys zs;
      Printf.fprintf oc "\tjump %s\n" x;
  | NonTail(a), CallCls(x, ys, zs) ->
      g'_args oc [(x, reg_cl)] ys zs;
      let ss = stacksize () in
      Printf.fprintf oc "\tstore %s %s %d\n" reg_ra reg_sp (ss - 1);
      Printf.fprintf oc "\taddi %s %s %d\n" reg_sp reg_sp ss;
      Printf.fprintf oc "\tjal min_caml_call_closure\n";
      Printf.fprintf oc "\tsubi %s %s %d\n" reg_sp reg_sp ss;
      Printf.fprintf oc "\tload %s %s %d\n" reg_ra reg_sp (ss - 1);
      if List.mem a allregs && a <> regs.(0) then
	Printf.fprintf oc "\taddi %s %s 0\n" a regs.(0)
  | NonTail(a), CallDir(Id.L(x), ys, zs) ->
      g'_args oc [] ys zs;
      let ss = stacksize () in
      let ra = Id.genid "ra" in
      Printf.fprintf oc "\tstore %s %s %d\n" reg_ra reg_sp (ss - 1);
      Printf.fprintf oc "\taddi %s %s %d\n" reg_sp reg_sp ss;
      Printf.fprintf oc "\tjal %s\n" x;
      Printf.fprintf oc "\tsubi %s %s %d\n" reg_sp reg_sp ss;
      Printf.fprintf oc "\tload %s %s %d\n" reg_ra reg_sp (ss - 1);
      if List.mem a allregs && a <> regs.(0) then
	Printf.fprintf oc "\taddi %s %s 0\n" a regs.(0)
  | _ -> assert false
and g'_tail_if oc e1 e2 l_else =
  let stackset_back = !stackset in
  g oc (Tail, e1);
  Printf.fprintf oc "%s:\n" l_else;
  stackset := stackset_back;
  g oc (Tail, e2)
and g'_non_tail_if oc dest e1 e2 l_else l_cont =
  let stackset_back = !stackset in
  g oc (dest, e1);
  let stackset1 = !stackset in
  Printf.fprintf oc "\tjump %s\n" l_cont;
  Printf.fprintf oc "%s:\n" l_else;
  stackset := stackset_back;
  g oc (dest, e2);
  Printf.fprintf oc "%s:\n" l_cont;
  Printf.fprintf oc "# this is debug message 2\n";
  let stackset2 = !stackset in
  stackset := S.inter stackset1 stackset2
and g'_args oc x_reg_cl ys zs =
  let (i, yrs) =
    List.fold_left
      (fun (i, yrs) y -> (i + 1, (y, regs.(i)) :: yrs))
      (0, x_reg_cl)
      ys in
  let (d, zfrs) =
    List.fold_left
      (fun (d, zfrs) z -> (d + 1, (z, fregs.(d)) :: zfrs))
      (i, [])
      zs in
  List.iter
    (fun (z, fr) ->
      Printf.fprintf oc "\taddi %s %s 0\n" fr z)
    (shuffle reg_sw (yrs @ zfrs))

let h oc { name = Id.L(x); args = _; fargs = _; body = e; ret = _ } =
  Printf.fprintf oc "%s:\n" x;
  stackset := S.empty;
  stackmap := [];
  g oc (Tail, e)

let f oc (Prog(fundefs, e)) =
  let fundefs = List.map
    (fun { name = name; args = args; fargs = fargs; body = e; ret = ret } ->
       { name = name; args = args; fargs = fargs; body = remove_percent e; ret = ret })
    fundefs
  in
  let e = remove_percent e in
  Format.eprintf "generating assembly...@.";
  Printf.fprintf oc ".section\t\".text\"\n";
  Printf.fprintf oc ".global\tmin_caml_start\n";
  Printf.fprintf oc "min_caml_start:\n";
  (*Printf.fprintf oc "\tsave\t%%sp, -112, %%sp\n"; (* from gcc; why 112? *)*)
  stackset := S.empty;
  stackmap := [];
  g oc (NonTail(regs.(0)), e);
  Printf.fprintf oc "\thlt\n";
  List.iter (fun fundef -> h oc fundef) fundefs;
(*
  Printf.fprintf oc ".section\t\".rodata\"\n";
  List.iter
    (fun (Id.L(x), d) ->
      Printf.fprintf oc "%s:\t# %f\n" x d;
      Printf.fprintf oc "\t.long\t0x%lx\n" (Int32.bits_of_float d))
    data;
*)
