open Printf
open Syntax

let linenum = ref 1

let rec string_of_syntax =
  let (~-) = string_of_syntax in
  function
  | Unit -> "()"
  | Bool b -> string_of_bool b
  | Int i -> string_of_int i
  | Float f -> string_of_float f
  | Not e -> sprintf "(not %s)" ~-e
  | Neg e -> sprintf "(~- %s)" ~-e
  | Add(e1,e2) -> sprintf "(+ %s\n%s)" ~-e1 ~-e2
  | Sub(e1,e2) -> sprintf "(- %s\n%s)" ~-e1 ~-e2
  | Mul(e1,e2) -> sprintf "(* %s\n%s)" ~-e1 ~-e2
  | Div(e1,e2) -> sprintf "(/ %s\n%s)" ~-e1 ~-e2
  | FNeg e -> sprintf "(~-. %s)" ~-e
  | FAdd(e1,e2) -> sprintf "(+. %s\n%s)" ~-e1 ~-e2
  | FSub(e1,e2) -> sprintf "(-. %s\n%s)" ~-e1 ~-e2
  | FMul(e1,e2) -> sprintf "(*. %s\n%s)" ~-e1 ~-e2
  | FDiv(e1,e2) -> sprintf "(/. %s\n%s)" ~-e1 ~-e2
  | Eq(e1,e2) -> sprintf "(= %s\n%s)" ~-e1 ~-e2
  | LE(e1,e2) -> sprintf "(<= %s\n%s)" ~-e1 ~-e2
  | If(e1,e2,e3) -> sprintf "(if %s\n%s\n%s)" ~-e1 ~-e2 ~-e3
  | Let((x,t),e1,e2) -> sprintf "(let (%s:%s %s)\n%s)"
      x (Id.id_of_typ t) ~-e1 ~-e2
  | Var x -> x
  | LetRec({ name = (name,t); args = argt; body = e1 }, e2) ->
      sprintf "(define (%s %s)\n%s)\n%s"
	name (Id.pp_list (List.map fst argt)) ~-e1 ~-e2
  | App(e1, es) -> sprintf "(%s %s)" ~-e1 (Id.pp_list (List.map (~-) es))
  | Tuple(es) -> sprintf "'(%s)" (Id.pp_list (List.map (~-) es))
  | LetTuple(xts, e1, e2) -> sprintf "(lett ((%s) %s)\n%s)"
      (Id.pp_list (List.map fst xts)) ~-e1 ~-e2
  | Array(e1,e2) -> sprintf "(newArray %s\n%s)" ~-e1 ~-e2
  | Get(e1,e2) -> sprintf "(%s)[%s]" ~-e1 ~-e2
  | Put(e1,e2,e3) -> sprintf "(setArray! %s\n%s\n%s)" ~-e1 ~-e2 ~-e3
