--------------------------------------------------------------------------------
-- Company: 
-- Engineer:
--
-- Create Date:    13:28:20 10/31/05
-- Design Name:    
-- Module Name:    cpu - Behavioral
-- Project Name:   
-- Target Device:  
-- Tool versions:  
-- Description:
--
-- Dependencies:
-- 
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
-- 
--------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_SIGNED.ALL;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
library UNISIM;
use UNISIM.VComponents.all;

entity cpu is
  port (
    CLK : in std_logic;
    LED : out std_logic_vector(7 downto 0);
    AA : out std_logic_vector(19 downto 0);
    ADVLDA : out std_logic;
    XRWA : out std_logic;
    CEA : out std_logic;
    BWA : out std_logic_vector(3 downto 0);
    CELA1X : out std_logic;
    CEHA1X : out std_logic;
    CEA2X : out std_logic;
    CEA2 : out std_logic;
    CLKMA0 : out std_logic;
    CLKMA1 : out std_logic;
    IOA : inout std_logic_vector(31 downto 0);
    -- IOPA : inout std_logic_vector(3 downto 0);
    LBOA : out std_logic;
    OEAX : out std_logic;
    ZZA : out std_logic;
    USBD : inout std_logic_vector(7 downto 0);
    USBRDX : out std_logic;
    USBWR : out std_logic;
    USBTXEX : in std_logic;
    USBRXFX : in std_logic;
    USBRSTX : out std_logic);

end cpu;

architecture Behavioral of cpu is

  component tick
    port (
      CLK    : in  std_logic;
      CLKFX  : out std_logic;
      CLKMA0 : out std_logic;
      CLKMA1 : out std_logic;
      LOCKED : out std_logic);
  end component;
  component heap
    port (
      CLK         : in     std_logic;
      LOCKED      : in     std_logic;
      enable      : in     std_logic;
      request     : in     std_logic_vector(1 downto 0);
      response    : out    std_logic_vector(1 downto 0);
      put_address : in     std_logic_vector(19 downto 0);
      get_address : out    std_logic_vector(19 downto 0);
      put_word    : in     std_logic_vector(31 downto 0);
      get_word    : out    std_logic_vector(31 downto 0);
      XRWA        : out    std_logic;
      AA          : out    std_logic_vector(19 downto 0);
      IOA         : inout  std_logic_vector(31 downto 0));
  end component;  
  component stack
    port (
      CLK               : in  std_logic;
      enable            : in  std_logic_vector(1 downto 0);
      put_address       : in  std_logic_vector(4 downto 0);
      put_word          : in  std_logic_vector(31 downto 0);
      get_left_address  : in  std_logic_vector(4 downto 0);
      get_left_word     : out std_logic_vector(31 downto 0);
      get_right_address : in  std_logic_vector(4 downto 0);
      get_right_word    : out std_logic_vector(31 downto 0));
  end component;
  
  component stream
    port (
      CLK            : in     std_logic;
      funct          : in     std_logic_vector(3 downto 0);
      put_left_word  : in     std_logic_vector(31 downto 0);
      put_right_word : in     std_logic_vector(31 downto 0);
      shift          : in     std_logic_vector(4 downto 0);
      get_comp_word  : out    std_logic_vector(31 downto 0);
      get_simp_word  : out    std_logic_vector(31 downto 0);
      complete       : out    std_logic;
      LED            : out    std_logic_vector(7 downto 0);
      USBD           : inout  std_logic_vector(7 downto 0);
      USBRDX         : out    std_logic;
      USBWR          : out    std_logic;
      USBTXEX        : in     std_logic;
      USBRXFX        : in     std_logic);
  end component;

  -- signal CLKIN : std_logic := '0';
  signal CLKMA : std_logic;
  signal RST : std_logic;
  signal LOCKED : std_logic;
  signal status : std_logic_vector(1 downto 0);
  signal hp_enable : std_logic;
  signal hp_request : std_logic_vector(1 downto 0);
  signal hp_response : std_logic_vector(1 downto 0);
  signal hp_put_address : std_logic_vector(19 downto 0);
  signal hp_get_address : std_logic_vector(19 downto 0);
  signal hp_put_word : std_logic_vector(31 downto 0);
  signal hp_get_word : std_logic_vector(31 downto 0);
  signal sk_enable : std_logic_vector(1 downto 0);
  signal sk_put_address : std_logic_vector(4 downto 0);
  signal sk_put_word : std_logic_vector(31 downto 0);
  signal sk_get_left_address : std_logic_vector(4 downto 0);
  signal sk_get_left_word : std_logic_vector(31 downto 0);
  signal sk_get_right_address : std_logic_vector(4 downto 0);
  signal sk_get_right_word : std_logic_vector(31 downto 0);
  signal sm_funct : std_logic_vector(3 downto 0);
  signal sm_put_left_word : std_logic_vector(31 downto 0);
  signal sm_put_right_word : std_logic_vector(31 downto 0);
  signal sm_shift : std_logic_vector(4 downto 0);
  signal sm_get_comp_word : std_logic_vector(31 downto 0);
  signal sm_get_simp_word : std_logic_vector(31 downto 0);
  signal sm_complete : std_logic;
  signal hp_response_buf : std_logic_vector(1 downto 0); 
  signal hp_get_address_buf : std_logic_vector(19 downto 0);
  signal hp_get_word_buf : std_logic_vector(31 downto 0);
  type sk_put_address_buf_t is array (2 downto 0) of std_logic_vector(4 downto 0);
  signal sk_put_address_buf : sk_put_address_buf_t;

begin

  ADVLDA <= '0';
  CEA <= '0';
  BWA <= "0000";
  CELA1X <= '0';
  CEHA1X <= '0';
  CEA2X <= '0';
  CEA2 <= '1';
  CLKMA0 <= CLKMA;
  CLKMA1 <= CLKMA;
  LBOA <= 'X';
  OEAX <= '0';
  ZZA <= '0';
  USBRSTX <= LOCKED;

  rc : ROC port map (
    O => RST);

--  process (CLK)
--  begin  -- process
--    if CLK'event and CLK = '1' then
--      CLKIN <= not CLKIN;
--    end if;
--  end process;

  cm: DCM
    port map(
      CLK0   => CLKMA,
      LOCKED => LOCKED,
      CLKFB  => CLKMA,
      CLKIN  => CLK,
      RST    => RST);  

  hp : heap port map (
    CLK         => CLK,
    LOCKED      => LOCKED,
    enable      => hp_enable,
    request     => hp_request,
    response    => hp_response,
    put_address => hp_put_address,
    get_address => hp_get_address,
    put_word    => hp_put_word,
    get_word    => hp_get_word,
    XRWA        => XRWA,
    AA          => AA,
    IOA         => IOA);

  sk : stack port map (
    CLK               => CLK,
    enable            => sk_enable,
    put_address       => sk_put_address,
    put_word          => sk_put_word,
    get_left_address  => sk_get_left_address,
    get_left_word     => sk_get_left_word,
    get_right_address => sk_get_right_address,
    get_right_word    => sk_get_right_word);

  sm : stream port map (
    CLK            => CLK,
    funct          => sm_funct,
    put_left_word  => sm_put_left_word,
    put_right_word => sm_put_right_word,
    shift          => sm_shift,
    get_comp_word  => sm_get_comp_word,
    get_simp_word  => sm_get_simp_word,
    complete       => sm_complete,
    LED            => LED,
    USBD           => USBD,
    USBRDX         => USBRDX,
    USBWR          => USBWR,
    USBTXEX        => USBTXEX,
    USBRXFX        => USBRXFX);

  sk_get_left_address <= hp_get_word(25 downto 21);
  sk_get_right_address <= hp_get_word(20 downto 16);

  status(1) <=
    '1' when sk_get_left_word = sk_get_right_word else
    '0' when sk_get_left_word /= sk_get_right_word else
    'X';

  status(0) <=
    '1' when sk_get_left_word < sk_get_right_word else
    '0' when sk_get_left_word >= sk_get_right_word else
    'X';

  process (CLK, LOCKED)
  begin
    if LOCKED = '0' then                -- asynchronous reset (active low)
      hp_get_address_buf <= "XXXXXXXXXXXXXXXXXXXX";
      hp_get_word_buf <= "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";      
      hp_response_buf <= "00";
    elsif CLK'event and CLK = '1' then  -- rising clock edge
      if sm_complete = '1' then
        hp_get_address_buf <= hp_get_address;
        hp_get_word_buf <= hp_get_word;
        hp_response_buf <= hp_response;        
      end if;
    end if;
  end process;

  process (status, hp_response_buf, hp_get_address_buf, hp_get_word_buf, sk_put_address_buf, sk_get_left_word, sk_get_right_word, sm_get_comp_word, sm_complete)
  begin  -- process
    case hp_response_buf is
      when "00" =>
        hp_enable <= sm_complete;
        hp_request <= "00";
        sk_enable <= '0' & sm_complete;
        sk_put_address <= "XXXXX";
        sk_put_word <= "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
        sm_funct <= "0000";
        sm_shift <= "XXXXX";
        sm_put_left_word <= "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
        sm_put_right_word <= "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
      when "01" =>
        case hp_get_word_buf(29 downto 26) is
          when "0000" =>
            hp_enable <= sm_complete;
            hp_request <= "00";
            sk_enable <= '1' & sm_complete;
            sk_put_address <= hp_get_word_buf(15 downto 11);
            sk_put_word <= sm_get_comp_word;
            sm_funct <=  hp_get_word_buf(3 downto 0);
            sm_shift <= hp_get_word_buf(10 downto 6);
            sm_put_left_word <= sk_get_left_word;
            sm_put_right_word <= sk_get_right_word;
          when "0001" =>
            hp_enable <= sm_complete;
            hp_request <= "00";
            sk_enable <= '1' & sm_complete;
            sk_put_address <= hp_get_word_buf(20 downto 16);
            sk_put_word <= sm_get_comp_word;
            sm_funct <= "0000";
            sm_shift <= "XXXXX";
            sm_put_left_word <= sk_get_left_word;
            sm_put_right_word <= sxt(hp_get_word_buf(15 downto 0), 32);
          when "0010" =>
            hp_enable <= sm_complete;
            hp_request <= "10";
            sk_enable <= '0' & sm_complete;
            sk_put_address <= "XXXXX";
            sk_put_word <= "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
            sm_funct <= "0000";
            sm_shift <= "XXXXX";
            sm_put_left_word <= sk_get_left_word;
            sm_put_right_word <= sxt(hp_get_word_buf(15 downto 0), 32);
          when "0011" =>
            hp_enable <= sm_complete;
            hp_request <= "11";
            sk_enable <= '0' & sm_complete;
            sk_put_address <= "XXXXX";
            sk_put_word <= "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
            sm_funct <= "0000";
            sm_shift <= "XXXXX";
            sm_put_left_word <= sk_get_left_word;
            sm_put_right_word <= sxt(hp_get_word_buf(15 downto 0), 32);
          when "1000" =>
            hp_enable <= sm_complete;
            hp_request <= '0' & status(1);
            sk_enable <= '0' & sm_complete;
            sk_put_address <= "XXXXX";
            sk_put_word <= "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
            sm_funct <= "0000";
            sm_shift <= "XXXXX";
            sm_put_left_word <= ext(hp_get_address_buf, 32);
            sm_put_right_word <= sxt(hp_get_word_buf(15 downto 0), 32);
          when "1001" =>
            hp_enable <= sm_complete;
            hp_request <= '0' & (not status(1));
            sk_enable <= '0' & sm_complete;
            sk_put_address <= "XXXXX";
            sk_put_word <= "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
            sm_funct <= "0000";
            sm_shift <= "XXXXX";
            sm_put_left_word <= ext(hp_get_address_buf, 32);
            sm_put_right_word <= sxt(hp_get_word_buf(15 downto 0), 32);
          when "1010" =>
            hp_enable <= sm_complete;
            hp_request <= '0' & status(0);
            sk_enable <= '0' & sm_complete;
            sk_put_address <= "XXXXX";
            sk_put_word <= "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
            sm_funct <= "0000";
            sm_shift <= "XXXXX";
            sm_put_left_word <= ext(hp_get_address_buf, 32);
            sm_put_right_word <= sxt(hp_get_word_buf(15 downto 0), 32);
          when "1011" =>
            hp_enable <= sm_complete;
            hp_request <= '0' & (status(0) or status(1));
            sk_enable <= '0' & sm_complete;
            sk_put_address <= "XXXXX";
            sk_put_word <= "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
            sm_funct <= "0000";
            sm_shift <= "XXXXX";
            sm_put_left_word <= ext(hp_get_address_buf, 32);
            sm_put_right_word <= sxt(hp_get_word_buf(15 downto 0), 32);
          when "1100" =>
            hp_enable <= sm_complete;
            hp_request <= '0' & ((sk_get_left_word(31) and sk_get_right_word(31)) xor status(0));
            sk_enable <= '0' & sm_complete;
            sk_put_address <= "XXXXX";
            sk_put_word <= "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
            sm_funct <= "0000";
            sm_shift <= "XXXXX";
            sm_put_left_word <= ext(hp_get_address_buf, 32);
            sm_put_right_word <= sxt(hp_get_word_buf(15 downto 0), 32);
          when "1101" =>
            hp_enable <= sm_complete;
            hp_request <= '0' & (((sk_get_left_word(31) and sk_get_right_word(31)) xor status(0)) or status(1));
            sk_enable <= '0' & sm_complete;
            sk_put_address <= "XXXXX";
            sk_put_word <= "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
            sm_funct <= "0000";
            sm_shift <= "XXXXX";
            sm_put_left_word <= ext(hp_get_address_buf, 32);
            sm_put_right_word <= sxt(hp_get_word_buf(15 downto 0), 32);
          when "1110" =>
            hp_enable <= sm_complete;
            hp_request <= "01";
            sk_enable <= '0' & sm_complete;
            sk_put_address <= "XXXXX";
            sk_put_word <= "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
            sm_funct <= "0000";
            sm_shift <= "XXXXX";
            sm_put_left_word <= sk_get_left_word;
            sm_put_right_word <= sxt(hp_get_word_buf(20 downto 0), 32);
          when "1111" =>
            hp_enable <= sm_complete;
            hp_request <= "01";
            sk_enable <= '1' & sm_complete;
            sk_put_address <= hp_get_word_buf(25 downto 21);
            sk_put_word <= "000000000000" & hp_get_address_buf;
            sm_funct <= "0000";
            sm_shift <= "XXXXX";
            sm_put_left_word <= ext(hp_get_address_buf, 32);
            sm_put_right_word <= sxt(hp_get_word_buf(20 downto 0), 32);
          when others =>
            hp_enable <= 'X';
            hp_request <= "XX";
            sk_enable <= "XX";
            sk_put_address <= "XXXXX";
            sk_put_word <= "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
            sm_funct <= "XXXX";
            sm_shift <= "XXXXX";
            sm_put_left_word <= "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
            sm_put_right_word <= "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
        end case;
      when "10" =>
        hp_enable <= sm_complete;
        hp_request <= "00";
        sk_enable <= '1'  & sm_complete;
        sk_put_address <= sk_put_address_buf(2);
        sk_put_word <= hp_get_word_buf;
        sm_funct <= "0000";
        sm_shift <= "XXXXX";
        sm_put_left_word <= "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
        sm_put_right_word <= "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
      when others =>
        hp_enable <= 'X';
        hp_request <= "XX";
        sk_enable <= "XX";
        sk_put_address <= "XXXXX";
        sk_put_word <= "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
        sm_funct <= "XXXX";
        sm_shift <= "XXXXX";
        sm_put_left_word <= "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
        sm_put_right_word <= "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
    end case;
  end process;

  hp_put_address <= sm_get_simp_word(19 downto 0);
  hp_put_word <= sk_get_right_word;
  
  process (CLK, LOCKED)
  begin
    if LOCKED = '0' then
      sk_put_address_buf <= ("XXXXX", "XXXXX", "XXXXX");
    elsif CLK'event and CLK = '1' then  -- rising clock edge
      if sm_complete = '1' then
        sk_put_address_buf <= sk_put_address_buf(1 downto 0) & hp_get_word_buf(20 downto 16);
      end if;
    end if;
  end process;

end Behavioral;
 
