--------------------------------------------------------------------------------
-- Company: 
-- Engineer:
--
-- Create Date:    09:04:32 12/05/05
-- Design Name:    
-- Module Name:    fpu - Behavioral
-- Project Name:   
-- Target Device:  
-- Tool versions:  
-- Description:
--
-- Dependencies:
-- 
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
-- 
--------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity fpu is
  port (
    CLK : in std_logic;
    funct : in  std_logic_vector(1 downto 0);
    put_left_word, put_right_word : in  std_logic_vector(31 downto 0);
    get_word : out  std_logic_vector(31 downto 0));
end fpu;

architecture Behavioral of fpu is
  
  procedure addsub (
    constant put_left_sign      : in  std_logic;
    constant put_right_sign     : in  std_logic;
    signal get_left_sign        : out std_logic;
    signal get_right_sign       : out std_logic;
    constant put_left_exponent  : in  std_logic_vector(8 downto 0);
    constant put_right_exponent : in  std_logic_vector(8 downto 0);
    signal get_exponent         : out std_logic_vector(8 downto 0);
    constant put_left_fraction  : in  std_logic_vector(24 downto 0);
    constant put_right_fraction : in  std_logic_vector(24 downto 0);
    signal get_left_fraction    : out std_logic_vector(24 downto 0);
    signal get_right_fraction   : out std_logic_vector(24 downto 0)) is
    variable left_shift : std_logic_vector(8 downto 0);
    variable right_shift : std_logic_vector(8 downto 0);
    variable shift_fraction : std_logic_vector(24 downto 0);
  begin
    left_shift := put_left_exponent - put_right_exponent;
    right_shift := put_right_exponent - put_left_exponent;
    if right_shift(8) & put_left_fraction >= left_shift(8) & put_right_fraction then
      get_left_sign <= put_left_sign;
      get_right_sign <= put_left_sign xor put_right_sign;
      get_exponent <= put_left_exponent;
      get_left_fraction <= put_left_fraction;
      if (left_shift(7) or left_shift(6) or left_shift(5)) = '0' then
        get_right_fraction <= shr(put_right_fraction, left_shift(4 downto 0));
      elsif (left_shift(7) or left_shift(6) or left_shift(5)) = '1' then
        get_right_fraction <= "0000000000000000000000000";
      else
        get_right_fraction <= "XXXXXXXXXXXXXXXXXXXXXXXXX";
      end if;
    else
      get_left_sign <= put_right_sign;
      get_right_sign <= put_left_sign xor put_right_sign;
      get_left_fraction <= put_right_fraction;
      get_exponent <= put_right_exponent;
      if (right_shift(7) or right_shift(6) or right_shift(5)) = '0' then
        get_right_fraction <= shr(put_left_fraction, right_shift(4 downto 0));
      elsif (right_shift(7) or right_shift(6) or right_shift(5)) = '1' then
        get_right_fraction <= "0000000000000000000000000";
      else
        get_right_fraction <= "XXXXXXXXXXXXXXXXXXXXXXXXX";
      end if;
    end if;
  end addsub;
    
  procedure mult (
    constant put_left_sign      : in  std_logic;
    constant put_right_sign     : in  std_logic;
    signal get_left_sign        : out std_logic;
    signal get_right_sign       : out std_logic;
    constant put_left_exponent  : in  std_logic_vector(8 downto 0);
    constant put_right_exponent : in  std_logic_vector(8 downto 0);
    signal get_exponent         : out std_logic_vector(8 downto 0);
    constant put_left_fraction  : in  std_logic_vector(24 downto 0);
    constant put_right_fraction : in  std_logic_vector(24 downto 0);
    signal get_left_fraction    : out std_logic_vector(24 downto 0);
    signal get_right_fraction   : out std_logic_vector(24 downto 0)) is
    variable shift_fraction : std_logic_vector(35 downto 0);
  begin
    get_left_sign <= put_left_sign xor put_right_sign;
    get_right_sign <= '0';
    get_exponent <= put_left_exponent + put_right_exponent - 128;
    shift_fraction := put_left_fraction(11 downto 0) * put_right_fraction(23 downto 0);
    get_left_fraction <= ext(shift_fraction(35 downto 23), 25);
    shift_fraction := put_left_fraction(23 downto 12) * put_right_fraction(23 downto 0);
    get_right_fraction <= shift_fraction(35 downto 11);
  end mult;
  
  signal left_sign, right_sign : std_logic;
  signal left_exponent, right_exponent : std_logic_vector(8 downto 0);
  signal left_fraction, right_fraction : std_logic_vector(24 downto 0);
  signal left_sign_buf, right_sign_buf : std_logic;
  signal exponent_buf : std_logic_vector(8 downto 0);
  signal left_fraction_buf, right_fraction_buf : std_logic_vector(24 downto 0);

begin

  process (put_left_word)
  begin  -- process
    left_sign <= put_left_word(31);
    left_exponent(8) <= '0';
    left_exponent(7 downto 0) <= put_left_word(30 downto 23);
    left_fraction(24) <= '0';
    left_fraction(23) <= put_left_word(30) or put_left_word(29) or put_left_word(28) or put_left_word(27) or put_left_word(26) or put_left_word(25) or put_left_word(24) or put_left_word(23);
    left_fraction(22 downto 0) <= put_left_word(22 downto 0);
  end process;

  process (put_right_word)
  begin  -- process
    right_sign <= put_right_word(31);
    right_exponent(8) <= '0';
    right_exponent(7 downto 0) <= put_right_word(30 downto 23);
    right_fraction(24) <= '0';
    right_fraction(23) <= put_right_word(30) or put_right_word(29) or put_right_word(28) or put_right_word(27) or put_right_word(26) or put_right_word(25) or put_right_word(24) or put_right_word(23);
    right_fraction(22 downto 0) <= put_right_word(22 downto 0);
  end process;

  process (CLK)
  begin  -- process
    if CLK'event and CLK = '1' then  -- rising clock edge
      case funct(1) is
        when '0' =>
          addsub(left_sign, right_sign xor funct(0), left_sign_buf, right_sign_buf, left_exponent, right_exponent, exponent_buf, left_fraction, right_fraction, left_fraction_buf, right_fraction_buf);
        when '1' =>
          mult(left_sign, right_sign xor funct(0), left_sign_buf, right_sign_buf, left_exponent, right_exponent, exponent_buf, left_fraction, right_fraction, left_fraction_buf, right_fraction_buf);
        when others =>
          left_sign_buf <= 'X';
          right_sign_buf <= 'X';
          exponent_buf <= "XXXXXXXXX";
          left_fraction_buf <= "XXXXXXXXXXXXXXXXXXXXXXXXX";
          right_fraction_buf <= "XXXXXXXXXXXXXXXXXXXXXXXXX";
      end case;
    end if;
  end process;

  process (left_sign_buf, right_sign_buf, exponent_buf, left_fraction_buf, right_fraction_buf)
    variable shift : std_logic_vector(4 downto 0);
    variable shift_exponent : std_logic_vector(8 downto 0);
    variable shift_fraction : std_logic_vector(24 downto 0);    
  begin  -- process
    if right_sign_buf = '0' then
      shift_fraction := left_fraction_buf + right_fraction_buf;
    elsif right_sign_buf = '1' then      
      shift_fraction := left_fraction_buf - right_fraction_buf;
    else      
      shift_fraction := "XXXXXXXXXXXXXXXXXXXXXXXXX";
    end if;
    shift(4) := not shift_fraction(24) and not shift_fraction(23) and not shift_fraction(22) and not shift_fraction(21) and not shift_fraction(20) and not shift_fraction(19) and not shift_fraction(18) and not shift_fraction(17) and not shift_fraction(16) and not shift_fraction(15) and not shift_fraction(14) and not shift_fraction(13) and not shift_fraction(12) and not shift_fraction(11) and not shift_fraction(10) and not shift_fraction(9);
    shift(3) := not shift_fraction(24) and not shift_fraction(23) and not shift_fraction(22) and not shift_fraction(21) and not shift_fraction(20) and not shift_fraction(19) and not shift_fraction(18) and not shift_fraction(17) and (shift_fraction(16) or shift_fraction(15) or shift_fraction(14) or shift_fraction(13) or shift_fraction(12) or shift_fraction(11) or shift_fraction(10) or shift_fraction(9) or (not shift_fraction(8) and not shift_fraction(7) and not shift_fraction(6) and not shift_fraction(5) and not shift_fraction(4) and not shift_fraction(3) and not shift_fraction(2) and not shift_fraction(1)));
    shift(2) := not shift_fraction(24) and not shift_fraction(23) and not shift_fraction(22) and not shift_fraction(21) and (shift_fraction(20) or shift_fraction(19) or shift_fraction(18) or shift_fraction(17) or (not shift_fraction(16) and not shift_fraction(15) and not shift_fraction(14) and not shift_fraction(13) and (shift_fraction(12) or shift_fraction(11) or shift_fraction(10) or shift_fraction(9) or (not shift_fraction(8) and not shift_fraction(7) and not shift_fraction(6) and not shift_fraction(5) and (shift_fraction(4) or shift_fraction(3) or shift_fraction(2) or shift_fraction(1))))));
    shift(1) := not shift_fraction(24) and not shift_fraction(23) and (shift_fraction(22) or shift_fraction(21) or (not shift_fraction(20) and not shift_fraction(19) and (shift_fraction(18) or shift_fraction(17) or (not shift_fraction(16) and not shift_fraction(15) and (shift_fraction(14) or shift_fraction(13) or (not shift_fraction(12) and not shift_fraction(11) and (shift_fraction(10) or shift_fraction(9) or (not shift_fraction(8) and not shift_fraction(7) and (shift_fraction(6) or shift_fraction(5) or (not shift_fraction(4) and not shift_fraction(3) and (shift_fraction(2) or shift_fraction(1))))))))))));
    shift(0) := not shift_fraction(24) and (shift_fraction(23) or (not shift_fraction(22) and (shift_fraction(21) or (not shift_fraction(20) and (shift_fraction(19) or (not shift_fraction(18) and (shift_fraction(17) or (not shift_fraction(16) and (shift_fraction(15) or (not shift_fraction(14) and (shift_fraction(13) or (not shift_fraction(12) and (shift_fraction(11) or (not shift_fraction(10) and (shift_fraction(9) or (not shift_fraction(8) and (shift_fraction(7) or (not shift_fraction(6) and (shift_fraction(5) or (not shift_fraction(4) and (shift_fraction(3) or (not shift_fraction(2) and (shift_fraction(1))))))))))))))))))))))));
    shift_exponent := exponent_buf - ("0" & (shift(4) and shift(3) and not exponent_buf(8) and not shift_fraction(0))& "00" & shift);
    shift_fraction := shl(shift_fraction, shift);
    if shift_exponent(8) = '0' then
      get_word <= left_sign_buf & (shift_exponent(7 downto 0) + 1) & shift_fraction(23 downto 1);
    elsif shift_exponent(8) = '1' then
      get_word <= left_sign_buf & "0000000000000000000000000000000";
    else
      get_word <= "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
    end if;    
  end process;

end Behavioral;
