--------------------------------------------------------------------------------
-- Company: 
-- Engineer:
--
-- Create Date:    19:18:24 11/09/05
-- Design Name:    
-- Module Name:    heap - Behavioral
-- Project Name:   
-- Target Device:  
-- Tool versions:  
-- Description:
--
-- Dependencies:
-- 
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
--
--------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity heap is
  port (
    CLK : in  std_logic;
    LOCKED : in std_logic;
    enable : in  std_logic;
    request : in  std_logic_vector(1 downto 0);
    response : out  std_logic_vector(1 downto 0);
    put_address : in  std_logic_vector(19 downto 0);
    get_address : out  std_logic_vector(19 downto 0);
    put_word : in  std_logic_vector(31 downto 0);
    get_word : out  std_logic_vector(31 downto 0);
    XRWA : out  std_logic;
    AA : out  std_logic_vector(19 downto 0);
    IOA : inout  std_logic_vector(31 downto 0));
end heap;

architecture Behavioral of heap is
  signal enable_buf : std_logic;
  signal request_buf : std_logic_vector(1 downto 0);
  signal response_buf : std_logic_vector(1 downto 0);
  signal put_address_buf, address, address_buf : std_logic_vector(19 downto 0);
  signal get_word_buf, put_word_buf : std_logic_vector(31 downto 0);
  signal request_buf_half : std_logic_vector(1 downto 0);
  signal put_word_buf_half : std_logic_vector(31 downto 0);
  
begin

  process (enable, request, address, put_address, put_address_buf, address_buf)
  begin  -- process
    if enable = '1' then
      case request is
        when "00" =>
          address <= put_address_buf + 1;
        when "01" =>
          address <= put_address;
        when "10" =>
          address <= put_address;
        when "11" =>
          address <= put_address;
        when others =>
          address <= "XXXXXXXXXXXXXXXXXXXX";
      end case;
    elsif enable = '0' then      
      address <= address_buf;
    else
      address <= "XXXXXXXXXXXXXXXXXXXX";
    end if;
  end process;

  XRWA <= request(1) nand request(0);
  AA <= address;
  
  process (CLK, LOCKED)
  begin  -- process
    if LOCKED = '0' then                -- asynchronous reset (active low)
      enable_buf <= '1';
      request_buf <= "11";
      put_address_buf <= "11111111111111111111";
      put_word_buf <= "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
      address_buf <= "XXXXXXXXXXXXXXXXXXXX";
    elsif CLK'event and CLK = '1' then  -- rising clock edge
      enable_buf <= enable;
      if enable = '1' then
        request_buf <= request;
      end if;
      if (enable and not request(1)) = '1' then
        put_address_buf <= address;
      end if;
      put_word_buf <= put_word;
      address_buf <= address;
    end if;
  end process;
  
  process (CLK, LOCKED)
  begin  -- process
    if LOCKED = '0' then                -- asynchronous reset (active low)
      response_buf <= "00";
      get_address <= "XXXXXXXXXXXXXXXXXXXX";
    elsif CLK'event and CLK = '1' then  -- rising clock edge
      if enable = '1' then
        response_buf <= (request_buf(1) and not request_buf(0)) & (not request_buf(1) and (request(1) or not request(0)));
        get_address <= address_buf;
      end if;
      request_buf_half <= request_buf;
      put_word_buf_half <= put_word_buf;
    end if;
    if LOCKED = '0' then                -- asynchronous reset (active low)
      IOA <= "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
    elsif CLK'event and CLK = '0' then  -- rising clock edge
      case request_buf_half is
        when "00" =>
          IOA <= "ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ";
        when "01" =>
          IOA <= "ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ";
        when "10" =>
          IOA <= "ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ";
        when "11" =>
          IOA <= put_word_buf_half;
        when others =>
          IOA <= "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
      end case;
    end if;
  end process;
  
  response <= response_buf(1) & (response_buf(0) and (request(1) or not request(0)));
  
  get_word <=
    IOA when enable_buf = '1' else
    get_word_buf when enable_buf = '0' else
    "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";

  process (CLK)
  begin  -- process
    if CLK'event and CLK = '1' then  -- rising clock edge
      if enable_buf = '1' then
        get_word_buf <= IOA;
      end if;
    end if;
  end process;  
  
end Behavioral;
