--------------------------------------------------------------------------------
-- Company: 
-- Engineer:
--
-- Create Date:    00:50:40 11/09/05
-- Design Name:    
-- Module Name:    stack - Behavioral
-- Project Name:   
-- Target Device:  
-- Tool versions:  
-- Description:
--
-- Dependencies:
-- 
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
-- 
--------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity stack is
  port (
    CLK : in  std_logic;
    enable : in  std_logic_vector(1 downto 0);
    put_address : in  std_logic_vector(4 downto 0);
    put_word : in  std_logic_vector(31 downto 0);
    get_left_address : in  std_logic_vector(4 downto 0);
    get_left_word : out  std_logic_vector(31 downto 0);
    get_right_address : in  std_logic_vector(4 downto 0);
    get_right_word : out  std_logic_vector(31 downto 0));
end stack;

architecture Behavioral of stack is
  type reg_t is array (31 downto 0) of std_logic_vector(31 downto 0);
  signal reg : reg_t := (others => "00000000000000000000000000000000");
  signal put_word_buf, get_left_word_buf, get_right_word_buf : std_logic_vector(31 downto 0);
  signal left_conflict, right_conflict : std_logic;

begin

  get_left_word <= put_word_buf when left_conflict = '1' else
                   get_left_word_buf when left_conflict = '0' else
                   "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
  get_right_word <= put_word_buf when right_conflict = '1' else
                    get_right_word_buf when right_conflict = '0' else
                   "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";

  process (CLK)
  begin  -- process
    if CLK'event and CLK = '1' then  -- rising clock edge
      if enable(1) = '1' then
        reg(conv_integer(put_address)) <= put_word;
      end if;
    end if;    
  end process;

  process (CLK)
  begin  -- process
    if CLK'event and CLK = '1' then  -- rising clock edge
      if enable(0) = '1' then
        put_word_buf <= put_word;
      end if;
    end if;    
  end process;

  process (CLK)
  begin  -- process
    if CLK'event and CLK = '1' then  -- rising clock edge
      if enable(0) = '1' then
        get_left_word_buf <= reg(conv_integer(get_left_address));
      end if;
    end if;
  end process;
  
  process (CLK)
  begin  -- process
    if CLK'event and CLK = '1' then  -- rising clock edge
      if enable(0) = '1' then
        get_right_word_buf <= reg(conv_integer(get_right_address));
      end if;
    end if;
  end process;
  
  process (CLK)
  begin  -- process
    if CLK'event and CLK = '1' then  -- rising clock edge
      if enable(0) = '1' then
        if get_left_address = put_address and enable = "11" then        
          left_conflict <= '1';
        elsif get_left_address /= put_address or enable /= "11" then        
          left_conflict <= '0';
        else
          left_conflict <= 'X';
        end if;
      end if;
    end if;
  end process;
  
  process (CLK)
  begin  -- process
    if CLK'event and CLK = '1' then  -- rising clock edge
      if enable(0) = '1' then
        if get_right_address = put_address and enable = "11" then
          right_conflict <= '1';
        elsif get_right_address /= put_address or enable /= "11" then        
          right_conflict <= '0';
        else
          right_conflict <= 'X';
        end if;
      end if;
    end if;
  end process;
  
end Behavioral;
