--------------------------------------------------------------------------------
-- Company: 
-- Engineer:
--
-- Create Date:    02:19:35 11/09/05
-- Design Name:    
-- Module Name:    stream - Behavioral
-- Project Name:   
-- Target Device:  
-- Tool versions:  
-- Description:
--
-- Dependencies:
-- 
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
-- 
--------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_SIGNED.ALL;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity stream is
  port (
    CLK : in std_logic;
    funct : in  std_logic_vector(3 downto 0);
    put_left_word, put_right_word : in  std_logic_vector(31 downto 0);
    shift : in  std_logic_vector(4 downto 0);
    get_comp_word, get_simp_word : out  std_logic_vector(31 downto 0);
    complete : out std_logic;
    LED : out std_logic_vector(7 downto 0);
    USBD : inout std_logic_vector(7 downto 0);
    USBRDX : out std_logic;
    USBWR : out std_logic;
    USBTXEX : in std_logic;
    USBRXFX : in std_logic);
end stream;

architecture Behavioral of stream is

  component au
    port (
      funct          : in  std_logic_vector(1 downto 0);
      put_left_word  : in  std_logic_vector(31 downto 0);
      put_right_word : in  std_logic_vector(31 downto 0);
      shift          : in  std_logic_vector(4 downto 0);
      get_word       : out std_logic_vector(31 downto 0));
  end component;

  component lu
    port (
      funct          : in  std_logic_vector(1 downto 0);
      put_left_word  : in  std_logic_vector(31 downto 0);
      put_right_word : in  std_logic_vector(31 downto 0);
      get_word       : out std_logic_vector(31 downto 0));
  end component;

  component fpu
    port (
      CLK            : in  std_logic;
      funct          : in  std_logic_vector(1 downto 0);
      put_left_word  : in  std_logic_vector(31 downto 0);
      put_right_word : in  std_logic_vector(31 downto 0);
     get_word       : out std_logic_vector(31 downto 0));
  end component;

  component xfpu
    port (
      CLK      : in  std_logic;
      count    : in  std_logic_vector(2 downto 0);
      funct    : in  std_logic_vector(1 downto 0);
      put_word : in  std_logic_vector(31 downto 0);
      get_word : out std_logic_vector(31 downto 0);
      LED      : out std_logic_vector(7 downto 0);
      USBD     : inout std_logic_vector(7 downto 0);
      USBRDX   : out std_logic;
      USBWR    : out std_logic);
  end component;
  
  signal sum : std_logic_vector(31 downto 0);
  signal complete_buf : std_logic;
  signal count : std_logic_vector(2 downto 0);
  signal a_funct : std_logic_vector(1 downto 0);
  signal a_put_left_word : std_logic_vector(31 downto 0);
  signal a_put_right_word : std_logic_vector(31 downto 0);
  signal a_shift : std_logic_vector(4 downto 0);
  signal a_get_word : std_logic_vector(31 downto 0);
  signal l_funct : std_logic_vector(1 downto 0);
  signal l_put_left_word : std_logic_vector(31 downto 0);
  signal l_put_right_word : std_logic_vector(31 downto 0);
  signal l_get_word : std_logic_vector(31 downto 0);
  signal f_funct : std_logic_vector(1 downto 0);
  signal f_put_left_word : std_logic_vector(31 downto 0);
  signal f_put_right_word : std_logic_vector(31 downto 0);
  signal f_get_word : std_logic_vector(31 downto 0);
  signal xf_count : std_logic_vector(2 downto 0);
  signal xf_funct : std_logic_vector(1 downto 0);
  signal xf_put_word : std_logic_vector(31 downto 0);
  signal xf_get_word : std_logic_vector(31 downto 0);

begin
    
  process (funct, put_left_word, put_right_word)
  begin  -- process
    if funct(0) = '0' then
      sum <= put_left_word + put_right_word;
    elsif funct(0) = '1' then
      sum <= put_left_word - put_right_word;
    else
      sum <= "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
    end if;
  end process;
  
  a_funct <= funct(1 downto 0);
  a_put_left_word <= put_left_word;
  a_put_right_word <= sum;
  a_shift <= shift;
  
  a : au port map (
    funct          => a_funct,
    put_left_word  => a_put_left_word,
    put_right_word => a_put_right_word,
    shift          => a_shift,
    get_word       => a_get_word);
  
  l_funct <= funct(1 downto 0);
  l_put_left_word <= put_left_word;
  l_put_right_word <= put_right_word;
  
  l : lu port map (
    funct          => l_funct,
    put_left_word  => l_put_left_word,
    put_right_word => l_put_right_word,
    get_word       => l_get_word);
  
  f_funct <= funct(1 downto 0);
  f_put_left_word <= put_left_word;
  f_put_right_word <= put_right_word;

  f : fpu port map (
    CLK            => CLK,
    funct          => f_funct,
    put_left_word  => f_put_left_word,
    put_right_word => f_put_right_word,
    get_word       => f_get_word);

  xf_count <= count;
  xf_funct <= funct(1 downto 0);
  xf_put_word <= put_left_word;
  
  xf : xfpu port map (
    CLK      => CLK,
    funct    => xf_funct,
    count    => xf_count,
    put_word => xf_put_word,
    get_word => xf_get_word,
    LED      => LED,
    USBD     => USBD,
    USBRDX   => USBRDX,
    USBWR    => USBWR);
    
  get_simp_word <= sum;
  
  process (funct, a_get_word, l_get_word, f_get_word, xf_get_word)
  begin  -- process
    case funct(3 downto 2) is
      when "00" =>
        get_comp_word <= a_get_word;
      when "01" =>
        get_comp_word <= l_get_word;
      when "10" =>
        get_comp_word <= f_get_word;
      when "11" =>
        get_comp_word <= xf_get_word;
      when others =>
        get_comp_word <= "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
    end case;
  end process;

  complete_buf <= not funct(3) or (not funct(2) and count(0)) or (not funct(1) and count(1)) or (count(2) and count(1) and count(0));

  complete <= complete_buf;

 process (CLK)
  begin  -- process
    if CLK'event and CLK = '1' then  -- rising clock edge
      if complete_buf = '1' then
        count <= "000";
      elsif (count /= "100" or funct(0) = '1' or USBRXFX = '0') and (count /= "010" or funct(0) = '0' or USBTXEX = '0') then
        count <= count + 1;
      end if;      
    end if;
  end process;


end Behavioral;
