--------------------------------------------------------------------------------
-- Company: 
-- Engineer:
--
-- Create Date:    18:37:36 12/14/05
-- Design Name:    
-- Module Name:    xfpu - Behavioral
-- Project Name:   
-- Target Device:  
-- Tool versions:  
-- Description:
--
-- Dependencies:
-- 
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
-- 
--------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
library UNISIM;
use UNISIM.VComponents.all;

entity xfpu is
  port (
    CLK : in std_logic;
    count : in  std_logic_vector(2 downto 0);
    funct : in  std_logic_vector(1 downto 0);
    put_word : in  std_logic_vector(31 downto 0);
    get_word : out  std_logic_vector(31 downto 0);
    LED : out std_logic_vector(7 downto 0);
    USBD : inout std_logic_vector(7 downto 0);
    USBRDX : out std_logic;
    USBWR : out std_logic);
end xfpu;

architecture Behavioral of xfpu is
    
  procedure div (
    constant put_word   : in  std_logic_vector(31 downto 0);    
    signal   get_word   : out std_logic_vector(31 downto 0);
    constant left_hint  : in  std_logic_vector(9 downto 0);
    constant right_hint : in  std_logic_vector(23 downto 0)) is
    variable shift_hint : std_logic_vector(34 downto 0);
  begin
    shift_hint := ('1' & left_hint) * ("000000000000000000000110" - right_hint);
    get_word(31) <= put_word(31);
    get_word(30 downto 23) <= not put_word(30 downto 23);
    get_word(22 downto 0) <= shift_hint(32 downto 10);
  end div;
    
  procedure sqrt (
    constant put_word   : in  std_logic_vector(31 downto 0);    
    signal   get_word   : out std_logic_vector(31 downto 0);
    constant left_hint  : in  std_logic_vector(23 downto 0);
    constant right_hint : in  std_logic_vector(23 downto 0)) is
    variable shift_hint : std_logic_vector(19 downto 0);
  begin
    if put_word(15) = '0' then
      shift_hint(19 downto 8) := left_hint(23 downto 12);
    elsif put_word(15) = '1' then
      shift_hint(19 downto 8) := left_hint(23 downto 12) + ext(left_hint(11 downto 8), 12) + 1;
    else
      shift_hint(19 downto 8) := "XXXXXXXXXXXX";
    end if;
    shift_hint(7 downto 0) := left_hint(7 downto 0);
    get_word(31) <= put_word(31);
    get_word(30) <= put_word(30);
    get_word(29) <= not put_word(30) and (put_word(29) or put_word(28) or put_word(27) or put_word(26) or put_word(25) or put_word(24) or put_word(23));
    get_word(28 downto 23) <= put_word(29 downto 24);
    get_word(22 downto 2) <= ('1' & shift_hint(19 downto 0))
                             + right_hint(22 downto 2);
    get_word(1 downto 0) <= right_hint(1 downto 0);
  end sqrt;
    
  signal left_key : std_logic_vector(9 downto 0);
  signal right_key : std_logic_vector(8 downto 0);
  signal left_hash : std_logic_vector(17 downto 0);
  signal right_hash : std_logic_vector(35 downto 0);
  signal product : std_logic_vector(33 downto 0);
  
begin
  
  process (funct, put_word)
  begin  -- process
    if funct(0) = '0' then
      left_key <= '0' & put_word(22 downto 14);
    elsif funct(0) = '1' then
      left_key <= '1' & put_word(22 downto 15) & put_word(23);
    else
      left_key <= "XXXXXXXXXX";
    end if;
    right_key <= '0' & put_word(22 downto 16) & put_word(23);
  end process;

  rb : RAMB16_S18_S36
    generic map (
INIT_00 => X"F11BF22AF2C0F3F4F4DBF5D5F6C0F7E4F89AF9BDFA80FB94FC9AFDA9FE80FF84",
INIT_01 => X"E25DE341E401E533E61CE6ECE7C1E8A2E99CEA97EB41EC53ED5CEE42EF01F003",
INIT_02 => X"D49ED55BD642D731D81ED8C5D982DAA2DB5DDC70DD02DE10DEDDDFDBE082E181",
INIT_03 => X"C79FC86FC903CA2CCADFCB9DCC43CD1FCE1FCEC7CF83D090D15ED231D2C3D3C0",
INIT_04 => X"BB61BC05BCC4BDAABE60BF2EBFC4C09AC160C21CC2C4C38CC460C546C603C6FC",
INIT_05 => X"AFA2B059B105B1E4B262B306B3C5B496B561B62FB6C5B788B861B91CB9C4BAB8",
INIT_06 => X"A4A3A52BA5C6A6A0A763A7D7A886A951AA23AA84AB46AC03ACE2ADACAE05AEF2",
INIT_07 => X"9A249ABE9B479BD89CA49D6A9DC79E8A9F249FD7A086A13DA1E4A283A306A3EF",
INIT_08 => X"9026908F914891D29265933B9388944394E595A6960796F79765981298C79967",
INIT_09 => X"86A7871C8789884B88E7894D8A088ABC8B268BF88C488D2F8DA68E238EC88F5F",
INIT_0A => X"7D687DEC7E8A7F037FA8801C80C9817381E88246830983A6842784F1854985D8",
INIT_0B => X"74E9757D75CA767976E9776777CA786B792979917A0A7A9E7B297BC17C4A7CD0",
INIT_0C => X"6C6B6CCB6D4B6DF06EAA6F346F8B702270AA7124718B721572AA730D738B7407",
INIT_0D => X"64AC6514658C6626666C66C3674C67D8686B68EC694C69CA6A6B6ADB6B4B6BFC",
INIT_0E => X"5CED5D605DCD5E5A5EED5F4E5FCD600C60AC6137618D620062AC6325638C6432",
INIT_0F => X"55AE562C568E5710576E57D5584E58C1596E59835A0D5AB55B2D5BB15C0D5C66",
INIT_10 => X"4EAF4F354F8F4FC3506F50E3514E51B7522F5291530E536853EE547E54CE551C",
INIT_11 => X"4830487F48CF493749B04A2C4A4F4AEB4B704B9A4C0F4C9C4D304D474DCF4E4F",
INIT_12 => X"41B241C4425042EA433143B143D0445D44F1451E459045D0467146CC47104784",
INIT_13 => X"3B733BCB3C113C9D3CF23D773D913E103E723EEB3F113F834032405740D04136",
INIT_14 => X"357435D33612368E36F3377F37923801387338F2391139B439F33A5E3A913B2A",
INIT_15 => X"2FB5301C305330C03135314831D2323332B432F4335233A933F434673492351B",
INIT_16 => X"2A362AA02AD32B302BB62BCC2C132CA52CF52D7E2D932DD82E752EAA2F132F4E",
INIT_17 => X"24F72525259425E42637269026D42716277727C32814288B28F6292E299329FD",
INIT_18 => X"1FB820252055209120F82157219521C72278228822D4233923B723F3241424AE",
INIT_19 => X"1AF91B2C1B561B831BF91C5E1C951CF51D391D881DD51E2A1E781EFA1F151F60",
INIT_1A => X"15FA166E169616F3173A175F17961824187A188918D6191919B919FB1A161A4E",
INIT_1B => X"117B11B111D7121F127B12E21317135413BB13D31417148914FA153D155615BE",
INIT_1C => X"0CFC0D350D580D900DFC0E250E980EC50F3C0F4F0F970FF6103C104010D7112B",
INIT_1D => X"08BD08F20918097E09BD09E30A180A730ABD0AD30B180B670BFD0C040C580C9C",
INIT_1E => X"047E04B804D90529057E05A805D90621067E069106D90716077E078107D9080A",
INIT_1F => X"007F00B800DA0119017F01A801DA020D027F029802DA0301037F038803D90435",
INIT_20 => X"64DFF8B8658AF9AA6678FA9E66EDFB9467A1FC8C6858FD866912FE82698FFF80",
INIT_21 => X"5F92F1276065F20960C9F32D61A2F4136209F4C162E8F5EB638AF6D76437F7C5",
INIT_22 => X"5A98EA395B2AEAD65BE8EBEA5C7EEC8B5D01EDA25D9BEEBC5E65EF615EC3F07F",
INIT_23 => X"55BFE33B5662E43D56E6E4D0576CE5A45817E6AC58A1E7435953E84F59E0E929",
INIT_24 => X"5100DC8A51B2DD4F5225DE1652F6DF0B536EDFD553E8E0A15482E19D5500E26B",
INIT_25 => X"4C8ED6694D30D7234D93D7DE4E39D89A4EF5D9584F5FDA174FCADB035078DBC6",
INIT_26 => X"4878D02E48CAD0DB495ED18A49F4D27A4A4CD32C4AE5D3DF4B41D4FB4C2ED5B1",
INIT_27 => X"442CCA5B44ADCB3D4531CBE045B7CC84463ECD6A4688CDEF4713CE9847A0CF42",
INIT_28 => X"4026C4A54097C57A410BC6114181C6EA4239C78442B2C800432EC8DD43ACC9BB",
INIT_29 => X"3C53BF7F3CF5BFF03D59C0BB3D8CC1473E34C1FB3E9DC28A3F09C35B3FB6C42D",
INIT_2A => X"38B8BA2B390ABA94399EBB5339C9BC143A60BC803AFABD433B56BDF13BF3BEB7",
INIT_2B => X"3539B53E355AB59F35DDB6523663B6F536EBB7AB3775B810379BB8C93828B970",
INIT_2C => X"31A9B02131C2B0B93275B16032EBB1FB3363B2A53380B302343BB3EF34B9B44E",
INIT_2D => X"2E0DAB5F2EAEABEF2F12AC8A2F65AD1D2FCCADB13076AE50308CAEE63139AF89",
INIT_2E => X"2AF0A6EF2B41A7782BD5A7C12C20A8922CB7A91E2D11A9AB2D5FAA382DFDAACF",
INIT_2F => X"27FDA246283EA2C72881A3492907A40D294BA49129D5A5162A1AA59C2AA8A667",
INIT_30 => X"24E89E2525209E9E25949F1825CF9F922646A00E2683A0CA273FA148277DA1C7",
INIT_31 => X"21F299C722239A7822879AEA22F99B5D23619BD023969C8024019D3624789DAD",
INIT_32 => X"1F0995E51F72964D1F9C96F720319762205D97CE20F7987B212598E921829958",
INIT_33 => X"1C1591C21C9692631CF892C51D7E93671DA293CB1DC794701E5294D61EB9957D",
INIT_34 => X"19BA8E0E19D38EA71A2D8F011AA28F9C1AFE90381B1B90951B9790C41BF59162",
INIT_35 => X"16DA8A55172B8AE6178E8B7817E18BCB18348C5F189E8C8A18F48D1F190B8DB6",
INIT_36 => X"145F86FF14A88748153C87AD154788381593888315E08910167B899E168A89C6",
INIT_37 => X"11F4835F1235838012B9840112FC8484130085081345856A138A85F014178677",
INIT_38 => X"0F5B7FF80FCD80711007808E1042810910BD818510FA82021176826111B582DF",
INIT_39 => X"0D067CA40D777CFC0DA87D6E0DDB7DE10E427E3B0EB77EAF0EED7F250F247F41",
INIT_3A => X"0AF779430B1F79970B497A010BB47AAC0BDF7AC20C797B6E0CA77BDC0CD67C34",
INIT_3B => X"08AF761608D076B7093176C80954776B09B877BD09DC78210A6878870A8F78DB",
INIT_3C => X"0670732C0689737906E373D2073D7420075874BB07D2751808307567084F75C6",
INIT_3D => X"044E7031049F707A04F170CC05047115055771A905AC71F305C2724906197294",
INIT_3E => X"021F6D2702686DB002B16DF502FB6E3A03076E8603A06ECD03EE6F5A043D6FA2",
INIT_3F => X"003F6A4F00406A8F00816AD000C46B5301076B95014C6C1901916C5C01D86CE3",
INITP_00 => X"97968496869796E69696869696D88686C886858685858878C7C7C887B777B777",
INITP_01 => X"55B8B5B8A558A5B7B555B7B745A544A444A5A4A59447A7A7A44696A797979497",
INITP_02 => X"626663D6655666526665C25656C65566C6C6566665C5C6B55665586555555855",
INITP_03 => X"637777777773F373736373736373E666676766636377E76373667662D6666666",
INITP_04 => X"3C42B29E89894C4220C612E4018BB406D881861FA770FB73A995D40C721D400C",
INITP_05 => X"736773E80875DCF84FB6DFACCBE655955584444CBBA4C51DDDCD191B6FA40FA7",
INITP_06 => X"F87663ADD95473B97366E472984CC50FAD8BA2D8473A611CFA1C70951C47E80C",
INITP_07 => X"C6E4C7F2A26B28FC3102DFA7E834F77772E907310EC4CBB996D58BB0F765C803")
    port map (
      DOA   => left_hash(17 downto 2),
      DOB   => right_hash(35 downto 4),
      DOPA  => left_hash(1 downto 0),
      DOPB  => right_hash(3 downto 0),
      ADDRA => left_key,
      ADDRB => right_key,
      CLKA  => CLK,
      CLKB  => CLK,
      DIA   => "XXXXXXXXXXXXXXXX",
      DIB   => "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
      DIPA  => "XX",
      DIPB  => "XXXX",
      ENA   => '1',
      ENB   => '1',
      SSRA  => '0',
      SSRB  => '0',
      WEA   => '0',
      WEB   => '0');

  process (CLK)
  begin
    if CLK'event and CLK = '1' then  -- rising clock edge
      product <= ('1' & put_word(22 downto 1)) * ('1' & left_hash(17 downto 8));
    end if;
  end process;
  
  process (funct, put_word, left_hash, right_hash, product, USBD)
  begin
    case funct is
      when "00" =>
        div(put_word, get_word, left_hash(17 downto 8), product(33 downto 10));
      when "01" =>
        sqrt(put_word, get_word, right_hash(25 downto 20) & right_hash(9 downto 0) & left_hash(7 downto 0), shl(ext(product(33 downto 11), 24), put_word(23 downto 23)));
      when "10" =>
        get_word <= "000000000000000000000000" & USBD;
      when "11" =>
        get_word <= "000000000000000000000000" & USBD;
      when others =>
        get_word <= "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
    end case;
  end process;
 
  process (CLK)
  begin  -- process
    if CLK'event and CLK = '1' then  -- rising clock edge
      if funct(0) = '0' then
        case count is
          when "000" => USBRDX <= '1';
          when "001" => USBRDX <= '1';
          when "010" => USBRDX <= '1';
          when "011" => USBRDX <= '1';
          when "100" => USBRDX <= '1';
          when "101" => USBRDX <= '0';
          when "110" => USBRDX <= '0';
          when "111" => USBRDX <= '0';
          when others => USBRDX <= 'X';        
        end case;
        USBWR <= '0';
      elsif funct(0) = '1' then        
        USBRDX <= '1';
        case count is
          when "000" => USBWR <= '0';
          when "001" => USBWR <= '0';
          when "010" => USBWR <= '0';
          when "011" => USBWR <= '1';
          when "100" => USBWR <= '1';
          when "101" => USBWR <= '1';
          when "110" => USBWR <= '0';
          when "111" => USBWR <= '0'; LED <= not put_word(7 downto 0);
          when others => USBWR <= 'X';        
        end case;
      else
        USBRDX <= 'X';
        USBWR <= 'X';
      end if;
    end if;
  end process;
  
  USBD <=
    "ZZZZZZZZ" when funct(0) = '0' else
    put_word(7 downto 0) when funct(0) = '1' else
    "XXXXXXXX";

end Behavioral;
