--------------------------------------------------------------------------------
-- Company: 
-- Engineer:
--
-- Create Date:    11:28:50 01/23/06
-- Design Name:    
-- Module Name:    accumulator - Behavioral
-- Project Name:   
-- Target Device:  
-- Tool versions:  
-- Description:
--
-- Dependencies:
-- 
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
-- 
--------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity accumulator is
  port (
    CLK                     : in  std_logic;
    LOCKED                  : in  std_logic;
    WE                      : in  std_logic;
    RAX, RAY, WA            : in  std_logic_vector (5 downto 0);
    RDX, RDY                : out std_logic_vector (31 downto 0);
    WD4, WD3, WD2, WD1, WD0 : in  std_logic_vector (31 downto 0);
    RVX, RVY                : out std_logic;
    WS3, WS2, WS1, WS0      : in  std_logic);
end accumulator;

architecture Behavioral of accumulator is

  signal E : std_logic_vector (4 downto 0);
  signal A : std_logic_vector (29 downto 0);
  signal D4, D3, D2, D1, D0, Q3, Q2, Q1, Q0 : std_logic_vector (31 downto 0);
  signal V3, V2, V1, V0, S3, S2, S1, S0 : std_logic;
  type ram_type is array (63 downto 0) of std_logic_vector (31 downto 0);
  signal RAM : ram_type;
  signal CX4, CX3, CX2, CX1, CX0, CY4, CY3, CY2, CY1, CY0 : boolean;
begin

  D0 <= WD0;
  
  with S0 select
    D1 <=
    WD1                                when '0',
    Q0                                 when '1',
    "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX" when others;

  with S1 select
    D2 <=
    WD2                                when '0',
    Q1                                 when '1',
    "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX" when others;
  
  with S2 select
    D3 <=
    WD3                                when '0',
    Q2                                 when '1',
    "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX" when others;
  
  with S3 select
    D4 <=
    WD4                                when '0',
    Q3                                 when '1',
    "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX" when others;
  
  
  V0 <= WS0;
  V1 <= S0 or WS1;
  V2 <= S1 or WS2;
  V3 <= S2 or WS3;

  RDX <= D0 when CX0 else
         D1 when CX1 else
         D2 when CX2 else
         D3 when CX3 else
         D4 when CX4 else
         RAM (conv_integer(RAX));
  
  RDY <= D0 when CY0 else
         D1 when CY1 else
         D2 when CY2 else
         D3 when CY3 else
         D4 when CY4 else
         RAM (conv_integer(RAY));

  RVX <= V0 when CX0 else
         V1 when CX1 else
         V2 when CX2 else
         V3 when CX3 else
         '1';

  RVY <= V0 when CY0 else
         V1 when CY1 else
         V2 when CY2 else
         V3 when CY3 else
         '1';
 
  CX4 <= E (4) = '1' and RAX = A (29 downto 24);
  CX3 <= E (3) = '1' and RAX = A (23 downto 18);
  CX2 <= E (2) = '1' and RAX = A (17 downto 12);
  CX1 <= E (1) = '1' and RAX = A (11 downto 6);
  CX0 <= E (0) = '1' and RAX = A (5 downto 0);
  CY4 <= E (4) = '1' and RAY = A (29 downto 24);
  CY3 <= E (3) = '1' and RAY = A (23 downto 18);
  CY2 <= E (2) = '1' and RAY = A (17 downto 12);
  CY1 <= E (1) = '1' and RAY = A (11 downto 6);
  CY0 <= E (0) = '1' and RAY = A (5 downto 0);
  
  process (CLK, LOCKED)
  begin  -- process
    if LOCKED = '0' then                -- asynchronous reset (active low)
      E <= "10000";
      A <= "000000XXXXXXXXXXXXXXXXXXXXXXXX";
      Q3 <= "00000000000000000000000000000000";
      Q2 <= "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
      Q1 <= "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
      Q0 <= "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
      S3 <= '1';
      S2 <= 'X';
      S1 <= 'X';
      S0 <= 'X';
    elsif CLK'event and CLK = '1' then  -- rising clock edge
      E <= E (3 downto 0) & WE;            
      A <= A (23 downto 0) & WA;
      Q3 <= D3;
      Q2 <= D2;
      Q1 <= D1;
      Q0 <= D0;
      S3 <= V3;
      S2 <= V2;
      S1 <= V1;
      S0 <= V0;
    end if;
  end process;

  process (CLK)
  begin  -- process
    if CLK'event and CLK = '1' then     -- falling clock edge
      if E(4) = '1' then
        RAM (conv_integer(A (29 downto 24))) <= D4;
      end if;
    end if;
  end process;
  
end Behavioral;
