--------------------------------------------------------------------------------
-- Company: 
-- Engineer:
--
-- Create Date:    11:29:28 01/23/06
-- Design Name:    
-- Module Name:    controller - Behavioral
-- Project Name:   
-- Target Device:  
-- Tool versions:  
-- Description:
--
-- Dependencies:
-- 
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
-- 
--------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity controller is
  port (
    CLK          : in  std_logic;
    LOCKED       : in  std_logic;    
    FD           : in  std_logic_vector(31 downto 0);
    FV           : in  std_logic;
    RVX, RVY     : in  std_logic;
    BS           : in  std_logic;
    FE           : out std_logic;
    WE           : out std_logic;
    RAX, RAY, WA : out std_logic_vector(5 downto 0);
    WS0          : out std_logic;
    LE           : out std_logic;
    TE           : out std_logic;    
    CND          : out std_logic_vector (2 downto 0);
    OPTD         : out std_logic_vector (5 downto 0);
    OPTS         : out std_logic;
    FNCT         : out std_logic_vector (3 downto 0);
    OFTD         : out std_logic_vector (15 downto 0);
    OFTS         : out std_logic;
    CNTS         : out std_logic);
end controller;

architecture Behavioral of controller is
  signal CV, RV, WV, LV, TV : std_logic;
  signal BV : std_logic_vector (2 downto 0);
begin

  RAX <= FD(27 downto 22);
  RAY <= FD(21 downto 16); 

  process (FD, RVX, RVY)
  begin
    case FD(31 downto 28) is
      when "0000" =>                    -- ADDI        
        RV <= RVX;
        WV <= '1';
        WA <= FD(21 downto 16);
        LV <= '0';
        TV <= '0';
        BV <= "111";
       when "0001" =>                    -- CALC
        RV <= RVX and RVY;
        WV <= '1';
        WA <= FD(15 downto 10);
        LV <= '0';
        TV <= '0';
        BV <= "111";
      when "0010" =>                    -- LOAD
        RV <= RVX;
        WV <= '1';
        WA <= FD(21 downto 16);
        LV <= '1';
        TV <= '0';
        BV <= "111";
      when "0011" =>                    -- STORE
        RV <= RVX and RVY;
        WV <= '0'; 
        WA <= "XXXXXX";
        LV <= '1';
        TV <= '1';
        BV <= "111";
      when "0100" =>                    -- BEI
        RV <= RVX;
        WV <= '0';
        WA <= "XXXXXX";
        LV <= '0';
        TV <= '0';
        BV <= "000";
      when "0101" =>                    -- BNEI
        RV <= RVX;
        WV <= '0'; 
        WA <= "XXXXXX";
        LV <= '0';
        TV <= '0';
        BV <= "001";
      when "0110" =>                    -- BLI
        RV <= RVX;
        WV <= '0'; 
        WA <= "XXXXXX";
        LV <= '0';
        TV <= '0';
        BV <= "010";
      when "0111" =>                    -- BGEI
        RV <= RVX;
        WV <= '0'; 
        WA <= "XXXXXX";
        LV <= '0';
        TV <= '0';
        BV <= "011";
      when "1000" =>                    -- BE
        RV <= RVX and RVY;
        WV <= '0'; 
        WA <= "XXXXXX";
        LV <= '0';
        TV <= '0';
        BV <= "000";
      when "1001" =>                    -- BNE
        RV <= RVX and RVY;
        WV <= '0'; 
        WA <= "XXXXXX";
        LV <= '0';
        TV <= '0';
        BV <= "001";
      when "1010" =>                    -- BL
        RV <= RVX and RVY;
        WV <= '0'; 
        WA <= "XXXXXX";
        LV <= '0';
        TV <= '0';
        BV <= "010";
      when "1011" =>                    -- BGE
        RV <= RVX and RVY;
        WV <= '0'; 
        WA <= "XXXXXX";
        LV <= '0';
        TV <= '0';
        BV <= "011";
      when "1100" =>                    -- FBL
        RV <= RVX and RVY;
        WV <= '0'; 
        WA <= "XXXXXX";
        LV <= '0';
        TV <= '0';
        BV <= "100";
      when "1101" =>                    -- FBGE
        RV <= RVX and RVY;
        WV <= '0'; 
        WA <= "XXXXXX";
        LV <= '0';
        TV <= '0';
        BV <= "101";
      when "1110" =>                    -- JAL
        RV <= RVX;
        WV <= '1';
        WA <= FD(21 downto 16);
        LV <= '0';
        TV <= '0';
        BV <= "110";
      when "1111" =>                    -- JR
        RV <= RVX and RVY;
        WV <= '0';
        WA <= "XXXXXX";
        LV <= '0';
        TV <= '0';
        BV <= "110";
      when others => null;
    end case;
  end process;

  CV <= RV and FV and not BS;
  FE <= RV or not FV;
  WE <= WV and CV;
  
  process (CLK, LOCKED)
  begin  -- process
    if LOCKED = '0' then                -- asynchronous reset (active low)
      LE <= '0';
      TE <= '0';
      CND <= "111";
      FNCT <= "XXXX";
      OPTD <= "XXXXXX";
      OFTD <= "XXXXXXXXXXXXXXXX";      
      WS0 <= '1';
      OPTS <= 'X';
      OFTS <= 'X';
      CNTS <= 'X';
    elsif CLK'event and CLK = '1' then  -- rising clock edge
      LE <= LV and CV;
      TE <= TV and CV;
      CND <= BV or (not CV, not CV, not CV);
      OPTD <= FD(21 downto 16);
      OFTD <= FD(15 downto 0);      
      case FD(31 downto 28) is
        when "0000" =>                  -- ADDI
          WS0 <= '1';
          OPTS <= 'X';
          OFTS <= '1';
          CNTS <= '0';
          FNCT <= "0000";
        when "0001" =>                  -- CALC
          WS0 <= not FD(3);
          OPTS <= 'X';
          OFTS <= '0';
          CNTS <= '0';
          FNCT <= FD (3 downto 0);
        when "0010" =>                  -- LOAD
          WS0 <= '0';
          OPTS <= 'X';
          OFTS <= '1';
          CNTS <= '0';
          FNCT <= "0XX0";
        when "0011" =>                  -- STORE
          WS0 <= 'X';
          OPTS <= 'X';
          OFTS <= '1';
          CNTS <= '0';
          FNCT <= "0XX0";
        when "0100" =>                  -- BEI
          WS0 <= 'X';
          OPTS <= '1';
          OFTS <= '1';
          CNTS <= '1';
          FNCT <= "0XX0";
        when "0101" =>                  -- BNEI
          WS0 <= 'X';
          OPTS <= '1';
          OFTS <= '1';
          CNTS <= '1';
          FNCT <= "0XX0";
        when "0110" =>                  -- BLI
          WS0 <= 'X';
          OPTS <= '1';
          OFTS <= '1';
          CNTS <= '1';
          FNCT <= "0XX0";
        when "0111" =>                  -- BGEI
          WS0 <= 'X';
          OPTS <= '1';
          OFTS <= '1';
          CNTS <= '1';
          FNCT <= "0XX0";
        when "1000" =>                  -- BE
          WS0 <= 'X';
          OPTS <= '0';
          OFTS <= '1';
          CNTS <= '1';
          FNCT <= "0XX0";
        when "1001" =>                  -- BNE
          WS0 <= 'X';
          OPTS <= '0';
          OFTS <= '1';
          CNTS <= '1';
          FNCT <= "0XX0";
        when "1010" =>                  -- BL
          WS0 <= 'X';
          OPTS <= '0';
          OFTS <= '1';
          CNTS <= '1';
          FNCT <= "0XX0";
        when "1011" =>                  -- BGE
          WS0 <= 'X';
          OPTS <= '0';
          OFTS <= '1';
          CNTS <= '1';
          FNCT <= "0XX0";
        when "1100" =>                  -- FBL
          WS0 <= 'X';
          OPTS <= '0';
          OFTS <= '1';
          CNTS <= '1';
          FNCT <= "0XX0";
        when "1101" =>                  -- FBGE
          WS0 <= 'X';
          OPTS <= '0';
          OFTS <= '1';
          CNTS <= '1';
          FNCT <= "0XX0";
        when "1110" =>                  -- JAL
          WS0 <= '1';
          OPTS <= 'X';
          OFTS <= '1';
          CNTS <= '1';
          FNCT <= "0XX0";
        when "1111" =>                  -- JR
          WS0 <= 'X';
          OPTS <= 'X';
          OFTS <= '1';
          CNTS <= '0';
          FNCT <= "0XX0";
        when others => null;
      end case;
    end if;
  end process;

end Behavioral;
