--------------------------------------------------------------------------------
-- Company: 
-- Engineer:
--
-- Create Date:    12:03:50 02/22/06
-- Design Name:    
-- Module Name:    fcalculator - Behavioral
-- Project Name:   
-- Target Device:  
-- Tool versions:  
-- Description:
--
-- Dependencies:
-- 
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
-- 
--------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
library UNISIM;
use UNISIM.VComponents.all;

entity fcalculator is
  port (
    CLK      : in  std_logic;
    LOCKED   : in  std_logic;
    RDX, RDY : in  std_logic_vector (31 downto 0);
    WD       : out std_logic_vector (31 downto 0);
    FNCT     : in  std_logic_vector (3 downto 0);
    CS       : in  std_logic;
    WS       : out std_logic);
end fcalculator;

architecture Behavioral of fcalculator is
  signal SGNDX, SGNDY, SGNQA, SGNQM, SGNQ, SGND : std_logic;
  signal EXPDX, EXPDY, EXPQA : std_logic_vector (7 downto 0);
  signal EXPQM, EXPQ, EXPD : std_logic_vector (8 downto 0);
  signal FRCDX, FRCDY, FRCQX, FRCQY, FRCQH, FRCQ : std_logic_vector (23 downto 0);
  signal FRCDL, FRCQLX, FRCQLY : std_logic_vector (12 downto 0);
  signal FRCDH, FRCDLX, FRCDLY : std_logic_vector (35 downto 0);
  signal FRCDA, FRCDM : std_logic_vector (24 downto 0);
  signal FRCD : std_logic_vector (22 downto 0);
  signal SHFTDX, SHFTDY : std_logic_vector (7 downto 0);
  signal SHFTQ : std_logic_vector (4 downto 0);
  signal S : std_logic_vector (2 downto 0);

begin

  SGNDX <= RDX(31);
  EXPDX(7 downto 0) <= RDX(30 downto 23);
  FRCDX(23) <= RDX(30) or RDX(29) or RDX(28) or RDX(27) or RDX(26) or RDX(25) or RDX(24) or RDX(23);
  FRCDX(22 downto 0) <= RDX(22 downto 0);

  SGNDY <= RDY(31);
  EXPDY(7 downto 0) <= RDY(30 downto 23);
  FRCDY(23) <= RDY(30) or RDY(29) or RDY(28) or RDY(27) or RDY(26) or RDY(25) or RDY(24) or RDY(23);
  FRCDY(22 downto 0) <= RDY(22 downto 0);

  SHFTDX <= EXPDY - EXPDX;
  SHFTDY <= EXPDY - EXPDX;

  MULT18X18HGH : MULT18X18 port map (
    P               => FRCDH,
    A(17 downto 11) => "0000001",
    A(10 downto 0)  => FRCDX(22 downto 12),
    B(17 downto 11) => "0000001",
    B(10 downto 0)  => FRCDY(22 downto 12));

  MULT18X18LOWX : MULT18X18 port map (
    P               => FRCDLX,
    A(17 downto 11) => "0000001",
    A(10 downto 0)  => FRCDX(22 downto 12),
    B(17 downto 12) => "000000",
    B(11 downto 0)  => FRCDY(11 downto 0));

  MULT18X18LOWY : MULT18X18 port map (
    P               => FRCDLY,
    A(17 downto 12) => "000000",
    A(11 downto 0)  => FRCDX(11 downto 0),
    B(17 downto 11) => "0000001",
    B(10 downto 0)  => FRCDY(22 downto 12));
  
  process (CLK, LOCKED)
  begin  -- process
    if LOCKED = '0' then                -- asynchronous reset (active low)
      S <= "XXX";
      SGNQA <= 'X';
      EXPQA <= "XXXXXXXX";
      FRCQX <= "XXXXXXXXXXXXXXXXXXXXXXXX";
      FRCQY <= "XXXXXXXXXXXXXXXXXXXXXXXX";
      SGNQM <= 'X';
      EXPQM <= "XXXXXXXXX";
      FRCQH <= "XXXXXXXXXXXXXXXXXXXXXXXX";
      FRCQLX <= "XXXXXXXXXXXXX";
      FRCQLY <= "XXXXXXXXXXXXX";
    elsif CLK'event and CLK = '1' then  -- rising clock edge
      S(2) <= FNCT(3) and not FNCT(2);
      S(1) <= FNCT(1);
      S(0) <= FNCT(0) xor SGNDX xor SGNDY;
      case CS is
        when '0' =>
          SGNQA <= SGNDX;
          EXPQA <= EXPDX;
          FRCQX <= FRCDX;
          case SHFTDY(7) or SHFTDY(6) or SHFTDY(5) or not FRCDY(23) is
            when '0' =>
              FRCQY <= shr("1" & FRCDY(22 downto 0), SHFTDY(4 downto 0));
            when '1' =>
              FRCQY <= "000000000000000000000000";
            when others => null;
          end case;          
        when '1' =>
          SGNQA <= SGNDY xor FNCT(0);
          EXPQA <= EXPDY;
          FRCQX <= FRCDY;
          case SHFTDX(7) or SHFTDX(6) or SHFTDX(5) or not FRCDX(23) is
            when '0' =>
              FRCQY <= shr("1" & FRCDX(22 downto 0), SHFTDX(4 downto 0));
            when '1' =>
              FRCQY <= "000000000000000000000000";
            when others => null;
          end case;
        when others => null;
      end case;
      SGNQM <= SGNDX xor SGNDY;
      EXPQM <= ("0" & EXPDX) + ("0" & EXPDY) - 128;
      FRCQH <= FRCDH(23 downto 0);
      FRCQLX <= FRCDLX(23 downto 11);
      FRCQLY <= FRCDLY(23 downto 11);
    end if;
  end process;

  FRCDL <= FRCQLX + FRCQLY + 1;
  FRCDM <= (FRCQH(23 downto 0) + FRCDL(12 downto 1) + FRCDL(0)) & not FRCDL(0);

  with S(0) select
    FRCDA <=
    ("0" & FRCQX) + ("0" & FRCQY)   when '0',
    ("0" & FRCQX) - ("0" & FRCQY)   when '1',
    "XXXXXXXXXXXXXXXXXXXXXXXXX"     when others;

  
  process (CLK, LOCKED)
  begin  -- process
    if LOCKED = '0' then                -- asynchrounous reset (active low)
      WS <= 'X';
      SGNQ <= 'X';
      EXPQ <= "XXXXXXXXX";
      FRCQ <= "XXXXXXXXXXXXXXXXXXXXXXXX";
      SHFTQ <= "XXXXX";
    elsif CLK'event and CLK = '1' then  -- rising clock edge
      WS <= S(2);
      case S(1) is
        when '0' =>
          SGNQ <= SGNQA;
          EXPQ <= "0" & EXPQA;
          FRCQ <= FRCDA(23 downto 0);
          SHFTQ(4) <= not FRCDA(24) and not FRCDA(23) and not FRCDA(22) and not FRCDA(21) and not FRCDA(20) and not FRCDA(19) and not FRCDA(18) and not FRCDA(17) and not FRCDA(16) and not FRCDA(15) and not FRCDA(14) and not FRCDA(13) and not FRCDA(12) and not FRCDA(11) and not FRCDA(10) and not FRCDA(9);
          SHFTQ(3) <= not FRCDA(24) and not FRCDA(23) and not FRCDA(22) and not FRCDA(21) and not FRCDA(20) and not FRCDA(19) and not FRCDA(18) and not FRCDA(17) and (FRCDA(16) or FRCDA(15) or FRCDA(14) or FRCDA(13) or FRCDA(12) or FRCDA(11) or FRCDA(10) or FRCDA(9) or (not FRCDA(8) and not FRCDA(7) and not FRCDA(6) and not FRCDA(5) and not FRCDA(4) and not FRCDA(3) and not FRCDA(2) and not FRCDA(1)));
          SHFTQ(2) <= not FRCDA(24) and not FRCDA(23) and not FRCDA(22) and not FRCDA(21) and (FRCDA(20) or FRCDA(19) or FRCDA(18) or FRCDA(17) or (not FRCDA(16) and not FRCDA(15) and not FRCDA(14) and not FRCDA(13) and (FRCDA(12) or FRCDA(11) or FRCDA(10) or FRCDA(9) or (not FRCDA(8) and not FRCDA(7) and not FRCDA(6) and not FRCDA(5) and (FRCDA(4) or FRCDA(3) or FRCDA(2) or FRCDA(1))))));
          SHFTQ(1) <= not FRCDA(24) and not FRCDA(23) and (FRCDA(22) or FRCDA(21) or (not FRCDA(20) and not FRCDA(19) and (FRCDA(18) or FRCDA(17) or (not FRCDA(16) and not FRCDA(15) and (FRCDA(14) or FRCDA(13) or (not FRCDA(12) and not FRCDA(11) and (FRCDA(10) or FRCDA(9) or (not FRCDA(8) and not FRCDA(7) and (FRCDA(6) or FRCDA(5) or (not FRCDA(4) and not FRCDA(3) and (FRCDA(2) or FRCDA(1))))))))))));
          SHFTQ(0) <= not FRCDA(24) and (FRCDA(23) or (not FRCDA(22) and (FRCDA(21) or (not FRCDA(20) and (FRCDA(19) or (not FRCDA(18) and (FRCDA(17) or (not FRCDA(16) and (FRCDA(15) or (not FRCDA(14) and (FRCDA(13) or (not FRCDA(12) and (FRCDA(11) or (not FRCDA(10) and (FRCDA(9) or (not FRCDA(8) and (FRCDA(7) or (not FRCDA(6) and (FRCDA(5) or (not FRCDA(4) and (FRCDA(3) or (not FRCDA(2) and (FRCDA(1))))))))))))))))))))))));
        when '1' =>
          SGNQ <= SGNQM;
          EXPQ <= EXPQM;
          FRCQ <= FRCDM(23 downto 0);
          SHFTQ <= "0000" & not FRCDM(24);
        when others => null;
      end case;
    end if;
  end process;

  SGND <= SGNQ;
  EXPD <= EXPQ - ("0" & (SHFTQ(4) and SHFTQ(3) and not EXPQ(8) and not FRCQ(0)) & "00" & SHFTQ);
  FRCD <= shl(FRCQ, SHFTQ)(23 downto 1);

  with EXPD(8) select
    WD <=
    SGND & EXPD(7 downto 0) & FRCD     when '0',
    "00000000000000000000000000000000" when '1',
    "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX" when others;    

end Behavioral;
