--------------------------------------------------------------------------------
-- Company: 
-- Engineer:
--
-- Create Date:    14:56:33 01/24/06
-- Design Name:    
-- Module Name:    fetcher - Behavioral
-- Project Name:   
-- Target Device:  
-- Tool versions:  
-- Description:
--
-- Dependencies:
-- 
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
-- 
--------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity fetcher is
  port (
    CLK    : in  std_logic;
    LOCKED : in  std_logic;
    FE, LE : in  std_logic;
    FA     : out std_logic_vector(19 downto 0);
    BA     : in  std_logic_vector(19 downto 0);
    FD     : out std_logic_vector(31 downto 0);
    FV     : out std_logic;
    BS     : in  std_logic;
    AA     : out std_logic_vector(19 downto 0);
    IOA    : in  std_logic_vector(31 downto 0));
end fetcher;

architecture Behavioral of fetcher is

  signal AX, AY : std_logic_vector (19 downto 0);
  signal BV, LV : std_logic;
  signal C : std_logic_vector (3 downto 0);
  type ram_type is array (3 downto 0) of std_logic_vector (31 downto 0);
  signal RAM : ram_type := (others => "00000000000000000000000000000000");
  signal A4, A3, A2, A1, A0 : std_logic_vector (19 downto 0);
  signal V4, V3, V2, V1, V0 : std_logic;
  signal D : std_logic_vector (31 downto 0);
  signal C3, C2, C1, C0 : std_logic_vector (1 downto 0);

begin

  FD <= D;  
  FA <= A4;
  FV <= V4 and not BV;
  
  AA <= A0;

  with BV or LV select
    A0 <=
    AX                     when '0',
    AY                     when '1',
    "XXXXXXXXXXXXXXXXXXXX" when others;
  V0 <= '1';
  C0 <= C1 + 1;


  process (CLK, LOCKED)
  begin  -- process
    if LOCKED = '0' then                -- asynchronous reset (active low)
      AX <= "XXXXXXXXXXXXXXXXXXXX";
      AY <= "00000000000000000000";
      BV <= '1';
      LV <= '0';
      C <= "XXXX";
      A4 <= "XXXXXXXXXXXXXXXXXXXX";
      A3 <= "XXXXXXXXXXXXXXXXXXXX";
      A2 <= "XXXXXXXXXXXXXXXXXXXX";
      A1 <= "XXXXXXXXXXXXXXXXXXXX";
      V4 <= 'X';
      V3 <= 'X';
      V2 <= 'X';
      V1 <= 'X';
      D <= "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
      C3 <= "XX";
      C2 <= "XX";
      C1 <= "00";
    elsif CLK'event and CLK = '1' then  -- rising clock edge
      if FE = '1' and LV /= '1' then
        AX <= A0 + 1;
      end if;
      AY <= BA;
      BV <= BS;
      LV <= LE;
      C <= C (1 downto 0) & C0;
      RAM (conv_integer(C (3 downto 2))) <= IOA;
      if FE = '1' then
        A4 <= A3;        
        V4 <= V3 and not BV;
        D <= RAM(conv_integer(C3));
      end if;
      if FE = '1' then
        A3 <= A2;        
        V3 <= V2 and not BV;
        C3 <= C2;
      end if;
      if FE = '1' then
        A2 <= A1;        
        V2 <= V1 and not BV;
        C2 <= C1;
      end if;
      if FE = '1' then
        A1 <= A0;
        V1 <= V0 and not LV;
        C1 <= C0;
      end if;
    end if;
  end process;

end Behavioral;
