--------------------------------------------------------------------------------
-- Company: 
-- Engineer:
--
-- Create Date:    10:36:05 01/23/06
-- Design Name:    
-- Module Name:    processor - Behavioral
-- Project Name:   
-- Target Device:  
-- Tool versions:  
-- Description:
--
-- Dependencies:
-- 
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
-- 
--------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
library UNISIM;
use UNISIM.VComponents.all;

entity processor is
  port (
    CLK    : in    std_logic;
    LED    : out   std_logic_vector (7 downto 0);
    AA     : out   std_logic_vector (19 downto 0);
    ADVLDA : out   std_logic;
    XRWA   : out   std_logic;
    CEA    : out   std_logic;
    BWA    : out   std_logic_vector (3 downto 0);
    CELA1X : out   std_logic;
    CEHA1X : out   std_logic;
    CEA2X  : out   std_logic;
    CEA2   : out   std_logic;
    CLKMA0 : out   std_logic;
    CLKMA1 : out   std_logic;
    IOA    : inout std_logic_vector (31 downto 0);
    LBOA   : out   std_logic;
    OEAX   : out   std_logic;
    ZZA    : out   std_logic);
    
end processor;

architecture Behavioral of processor is

  component fetcher
    port (
      CLK    : in  std_logic;
      LOCKED : in  std_logic;
      FE, LE : in  std_logic;
      FA     : out std_logic_vector (19 downto 0);
      BA     : in  std_logic_vector (19 downto 0);
      FD     : out std_logic_vector (31 downto 0);
      FV     : out std_logic;
      BS     : in  std_logic;
      AA     : out std_logic_vector (19 downto 0);
      IOA    : in  std_logic_vector (31 downto 0));
  end component;

  component accumulator
    port (
      CLK                     : in  std_logic;
      LOCKED                  : in  std_logic;
      WE                      : in  std_logic;
      RAX, RAY, WA            : in  std_logic_vector (5 downto 0);
      RDX, RDY                : out std_logic_vector (31 downto 0);
      WD4, WD3, WD2, WD1, WD0 : in  std_logic_vector (31 downto 0);
      RVX, RVY                : out std_logic;
      WS3, WS2, WS1, WS0      : in  std_logic);
  end component;
  
  component controller
    port (
      CLK          : in  std_logic;
      LOCKED       : in  std_logic;
      FD           : in  std_logic_vector (31 downto 0);
      FV           : in  std_logic;
      RVX, RVY     : in  std_logic;
      BS           : in  std_logic;
      FE           : out std_logic;
      WE           : out std_logic;
      RAX, RAY, WA : out std_logic_vector (5 downto 0);
      WS0          : out std_logic;
      LE, TE       : out std_logic;
      CND          : out std_logic_vector (2 downto 0);
      OPTD         : out std_logic_vector (5 downto 0);
      OPTS         : out std_logic;
      FNCT         : out std_logic_vector (3 downto 0);
      OFTD         : out std_logic_vector (15 downto 0);
      OFTS         : out std_logic;
      CNTS         : out std_logic);
  end component;

  component comparator
    port (
      CND      : in  std_logic_vector (2 downto 0);
      RDX, RDY : in  std_logic_vector (31 downto 0);
      BS       : out std_logic;
      CS       : out std_logic;
      OPTD     : in  std_logic_vector (5 downto 0);
      OPTS     : in  std_logic);
  end component;

  component calculator
    port (
      RDX, RDY : in  std_logic_vector (31 downto 0);
      WD0      : out std_logic_vector (31 downto 0);
      FA       : in  std_logic_vector (19 downto 0);
      BA       : out std_logic_vector (19 downto 0);
      FNCT     : in  std_logic_vector (2 downto 0);
      OFTD     : in  std_logic_vector (15 downto 0);
      OFTS     : in  std_logic;
      CNTS     : in  std_logic);
  end component;

  component fcalculator
    port (
      CLK      : in  std_logic;
      LOCKED   : in  std_logic;
      RDX, RDY : in  std_logic_vector (31 downto 0);
      WD       : out std_logic_vector (31 downto 0);
      FNCT     : in  std_logic_vector (3 downto 0);
      CS       : in  std_logic;
      WS       : out std_logic);
  end component;
  
  component loader
    port (
      CLK    : in    std_logic;
      LOCKED : in    std_logic;
      LE     : in    std_logic;
      TE     : in    std_logic;
      LA     : in    std_logic_vector (19 downto 0);
      WD4    : out   std_logic_vector (31 downto 0);
      RD     : in    std_logic_vector (31 downto 0);
      XRWA   : out   std_logic;
      IOA    : inout std_logic_vector (31 downto 0));
  end component;


  signal CLK0, CLK2X, CLK2X180 : std_logic;
  signal RST : std_logic;
  signal LOCKED, LOCKEDG : std_logic;
  signal FE : std_logic;
  signal FA, BA : std_logic_vector (19 downto 0);
  signal BS, CS : std_logic;
  signal FD : std_logic_vector (31 downto 0);
  signal FV : std_logic;
  signal WE : std_logic;
  signal RAX, RAY, WA : std_logic_vector (5 downto 0);
  signal RDX, RDY, WD4, WD3, WD2, WD1, WD0, DX, DY : std_logic_vector (31 downto 0);
  signal RVX, RVY, WS3, WS2, WS1, WS0 : std_logic;
  signal LE, TE : std_logic;
  signal CND : std_logic_vector (2 downto 0);
  signal OPTD : std_logic_vector (5 downto 0);
  signal OPTS : std_logic;
  signal FNCT : std_logic_vector (3 downto 0);
  signal OFTD : std_logic_vector (15 downto 0);
  signal OFTS, CNTS : std_logic;

begin

  WS3 <= '0';
  WS1 <= '0';

  LED <= not WD0(7 downto 0);
  ADVLDA <= '0';
  CEA <= '0';
  BWA <= "0000";
  CELA1X <= '0';
  CEHA1X <= '0';
  CEA2X <= '0';
  CEA2 <= '1';
  CLKMA0 <= CLK2X;
  CLKMA1 <= CLK2X;
  LBOA <= 'X';
  OEAX <= '0';
  ZZA <= '0';

  RC : roc
    port map (
      O => RST);

  process (CLK2X180, LOCKED)
  begin
    if LOCKED = '0' then
      LOCKEDG <= '0';
    elsif CLK2X180'event and CLK2X180 = '1' then
      LOCKEDG <= '1';
    end if;
  end process;
  
  cm: DCM  port map(
    CLK0     => CLK0,
    CLK2X    => CLK2X,
    CLK2X180 => CLK2X180,
    LOCKED   => LOCKED,
    CLKFB    => CLK0,
    CLKIN    => CLK,
    RST      => RST);

  FETCH : fetcher port map (
    CLK    => CLK2X180,
    LOCKED => LOCKEDG,
    FE     => FE,
    LE     => LE,
    FA     => FA,
    BA     => BA,
    FD     => FD,
    FV     => FV,
    BS     => BS,
    AA     => AA,
    IOA    => IOA);

  ACCUM : accumulator port map (
    CLK    => CLK2X180,
    LOCKED => LOCKEDG,
    WE     => WE,
    RAX    => RAX,
    RAY    => RAY,
    WA     => WA,
    RDX    => RDX,
    RDY    => RDY,
    WD4    => WD4,
    WD3    => WD3,
    WD2    => WD2,
    WD1    => WD1,
    WD0    => WD0,
    RVX    => RVX,
    RVY    => RVY,
    WS3    => WS3,
    WS2    => WS2,
    WS1    => WS1,
    WS0    => WS0);

  process (CLK2X180, LOCKEDG)
  begin  -- process
    if LOCKEDG = '0' then                -- asynchronous reset (active low)
      DX <= "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
      DY <= "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
    elsif CLK2X180'event and CLK2X180 = '1' then     -- rising clock edge
      DX <= RDX;
      DY <= RDY;
    end if;
  end process;
  
  CTRL : controller port map (
    CLK    => CLK2X180,
    LOCKED => LOCKEDG,
    FD     => FD,
    FV     => FV,
    RVX    => RVX,
    RVY    => RVY,
    BS     => BS,
    FE     => FE,
    WE     => WE,
    RAX    => RAX,
    RAY    => RAY,
    WA     => WA,
    WS0    => WS0,
    LE     => LE,
    TE     => TE,
    CND    => CND,
    OPTD   => OPTD,
    OPTS   => OPTS,
    FNCT   => FNCT,
    OFTD   => OFTD,
    OFTS   => OFTS,
    CNTS   => CNTS);

  CMP : comparator port map (
    CND  => CND,
    RDX  => DX,
    RDY  => DY,
    BS   => BS,
    CS   => CS,
    OPTD => OPTD,
    OPTS => OPTS);

  CALC : calculator port map (
    RDX  => DX,
    RDY  => DY,
    WD0  => WD0,
    FA   => FA,
    BA   => BA,
    FNCT => FNCT(2 downto 0),
    OFTD => OFTD,
    OFTS => OFTS,
    CNTS => CNTS);

  FCALC : fcalculator port map (
    CLK    => CLK2X180,
    LOCKED => LOCKED,
    RDX    => DX,
    RDY    => DY,
    WD     => WD2,
    FNCT   => FNCT,
    CS     => CS,
    WS     => WS2);

  LOAD : loader port map (
    CLK    => CLK2X180,
    LOCKED => LOCKEDG,
    LE     => LE,
    TE     => TE,
    LA     => BA,
    RD     => DY,
    WD4    => WD4,
    XRWA   => XRWA,
    IOA    => IOA);


end Behavioral;
